"""
Analysis
--------
The ``maica_old.util.analysis`` module contains useful functions to analyze prediction results
of the machine learning algorithms.
"""


import numpy


def get_target_dist(targets: numpy.ndarray,
                    n_bins: int = 10):
    """
    Get a histogram of the target values in ``targets``. It returns bins and labels of the histogram.
    This function is useful to analyze the prediction results based on the target distribution for each range.

    :param targets: (*numpy.ndarray*) Target values of the dataset.
    :param n_bins: (*int*) The number of bins in the histogram (*default* = 10).
    :return: Lists of the bins and the labels of the histogram.
    """

    __targets = targets.flatten()
    bins = numpy.zeros(n_bins)

    # Get labels of the bins.
    targets_min = numpy.min(__targets)
    targets_max = numpy.max(__targets)
    label_bins = numpy.linspace(targets_min, targets_max, n_bins + 1)

    # Compute the number of data for each bin.
    for i in range(0, __targets.shape[0]):
        included = False

        for j in range(0, label_bins.shape[0] - 1):
            if label_bins[j] <= __targets[i] < label_bins[j + 1]:
                bins[j] += 1
                included = True
                break

        # If the target data belongs to the last bin.
        if not included:
            bins[n_bins - 1] += 1

    return bins.tolist(), label_bins.tolist()


def get_error_dist(targets: numpy.ndarray,
                   preds: numpy.ndarray,
                   n_ranges: int = 10):
    """
    Compute mean of prediction errors for each range of the target values in ``targets``.
    The prediction error is calculated by mean absolute error (MAE) between ``targets`` and ``preds``.
    This function is useful to analyze the prediction accuracy for each range of the target values.
    The prediction error of the :math:`k^{th}` range is calculated by:

    .. math::
        error_k = \\frac{1}{|S_k|} \sum_{i \in S_K} |y_i - y_i^{'}|,

    where :math:`S_k` is a set of indices of the data in the :math:`k^{th}` range,
    and :math:`y_i^{'}` is the predicted value of the input data :math:`\mathbf{x}_i`.

    :param targets: (*numpy.ndarray*) Target values of the prediction.
    :param preds: (*numpy.ndarray*) Predicted values.
    :param n_ranges: (*int*) The number of sections to split the range of the target values (*default* = 10).
    :return: Mean of errors and labels for each target range.
    """

    __targets = targets.flatten()
    __preds = preds.flatten()
    errors = numpy.zeros(n_ranges)
    bins = numpy.zeros(n_ranges)

    # Calculate ranges of bins.
    targets_min = numpy.min(__targets)
    targets_max = numpy.max(__targets)
    labels = numpy.linspace(targets_min, targets_max, n_ranges + 1)

    # Compute the error and the number of data for each bin.
    for i in range(0, __targets.shape[0]):
        included = False

        for j in range(0, labels.shape[0] - 1):
            if labels[j] <= __targets[i] < labels[j + 1]:
                errors[j] += numpy.abs(__targets[i] - __preds[i])
                bins[j] += 1
                included = True
                break

        # If the target data belongs to the last bin.
        if not included:
            errors[n_ranges - 1] += numpy.abs(__targets[i] - __preds[i])
            bins[n_ranges - 1] += 1

    # Calculate the mean of the errors for each bin.
    for i in range(0, n_ranges):
        if bins[i] == 0:
            errors[i] = 0
        else:
            errors[i] /= bins[i]

    return errors.tolist(), labels.tolist()



"""
Molecular Structure
-------------------
This module provides abstracted functions to convert the molecular structures into the mathematical graphs.
The converted molecular structure is stored as a ``torch_geometric.Data`` object.
"""


import torch
from torch_geometric.data import Data
from rdkit import Chem
from maica_old.chem.base import *
from maica_old.data.util import get_one_hot_feat


def get_mol_graph(elem_feats: numpy.ndarray,
                  smiles: str,
                  numeric_feats: numpy.ndarray,
                  target: float = None,
                  gid: int = -1,
                  present_hydrogen=False):
    """
    Convert the SMILES representation of a molecular structure into the mathematical graph
    :math:`G = (\mathcal{V}, \mathcal{E}, X, R)`, where :math:`\mathcal{V}` is a set of atoms,
    :math:`\mathcal{E}` is a set of bonds, :math:`X` is a atom-feature matrix,
    and :math:`R` is a bond-feature matrix.

    :param elem_feats: (*numpy.ndarray*) The NumPy array of elemental features.
    :param smiles: (*str*) A SMILES representation of the input molecular structure.
    :param numeric_feats: (*numpy.ndarray*) Numerical features of the molecule (e.g., molecular weight).
    :param target: (*float, optional*) The target property that you want to predict from the molecular structure (*default* = ``None``).
    :param gid: (*int, optional*) An integer identifier of the molecular structure (*default* = -1).
    :param present_hydrogen: (*bool, optional*) An argument to present hydrogens in the molecular graph (*default* = ``False``).
    :return: (*torch_geometric.Data*) A molecular graph :math:`G = (\mathcal{V}, \mathcal{E}, X, R)`.
    """

    try:
        # Get RDKit.Mol object from the given SMILES.
        mol = Chem.MolFromSmiles(smiles)

        # Present hydrogens in the molecule object.
        if present_hydrogen:
            mol = Chem.AddHs(mol)

        if mol is None:
            return None

        # Global information of the molecule.
        n_rings = mol.GetRingInfo().NumRings()

        atom_feats = list()
        bonds = list()
        bond_feats = list()

        # Generate node-feature matrix.
        for atom in mol.GetAtoms():
            # Get elemental features of the atom.
            elem_attr = elem_feats[atom.GetAtomicNum() - 1, :]

            # Get hybridization type of the atom.
            hbd_type = get_one_hot_feat(cat_hbd, str(atom.GetHybridization()))

            # Get formal charge of the atom.
            fc_type = get_one_hot_feat(cat_fc, str(atom.GetFormalCharge()))

            # Check whether the atom belongs to the aromatic ring in the molecule.
            mem_aromatic = 1 if atom.GetIsAromatic() else 0

            # Get the number of bonds.
            degree = atom.GetDegree()

            # Get the number of hydrogen bonds.
            n_hs = atom.GetTotalNumHs()

            # Concatenate the given numerical features if they exist.
            if numeric_feats is None:
                atom_feats.append(numpy.hstack([elem_attr, hbd_type, fc_type, mem_aromatic,
                                                degree, n_hs, n_rings]))
            else:
                atom_feats.append(numpy.hstack([elem_attr, hbd_type, fc_type, mem_aromatic,
                                                degree, n_hs, n_rings, numeric_feats]))

        # Check bonds in the molecule.
        for bond in mol.GetBonds():
            bonds.append([bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()])
            bond_feats.append(get_one_hot_feat(cat_bond_types, str(bond.GetBondType())))

            bonds.append([bond.GetEndAtomIdx(), bond.GetBeginAtomIdx()])
            bond_feats.append(get_one_hot_feat(cat_bond_types, str(bond.GetBondType())))

        # Check isolated graph.
        if len(bonds) == 0:
            raise RuntimeError

        # Convert numpy.ndarray to torch.Tensor.
        atom_feats = torch.tensor(atom_feats, dtype=torch.float)
        bonds = torch.tensor(bonds, dtype=torch.long).t().contiguous()
        bond_feats = torch.tensor(bond_feats, dtype=torch.float)
        gid = torch.tensor(gid, dtype=torch.long).view(1, 1)

        # Check target property.
        if target is not None:
            target = torch.tensor(target, dtype=torch.float).view(1, 1)

        return Data(x=atom_feats, edge_index=bonds, edge_attr=bond_feats, y=target, gid=gid)
    except RuntimeError:
        return None



"""
Machine Learning Utilities
--------------------------
The ``maica.ml.util`` module provides essential functions for training configuration and model reuse.
Most deep learning algorithms in MAICA are based on this module.
"""

import numpy
import copy
import torch
import pandas
import os
import graphviz
import xgboost
import maica.ml.base
import maica.data.base
import torch.utils.data as tdata
import torch_geometric.data as tgdata
import maica.core.sys as sys
import maica.core.env as env
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_percentage_error
from sklearn.metrics import r2_score
from maica.core.env import *
from sklearn.tree import export_graphviz
from maica.ml.base import SKLearnModel
from maica.data.base import Dataset
from maica.data.vector import VectorDataset
from maica.data.formula import FormDataset
from maica.data.graph import GraphDataset


class KFoldGenerator:
    def __init__(self,
                 dataset: Dataset,
                 k: int):
        self.sub_datasets = dataset.get_sub_datasets(k=k)

    def get(self,
            idx_fold: int):
        """
        Get training and test datasets for a given ``idx_fold``.

        :param idx_fold: (*int*) An index of the k-fold dataset.
        :return: Training and test datasets.
        """

        list_train_datasets = [self.sub_datasets[i] for i in range(0, len(self.sub_datasets)) if i != idx_fold]
        train_dataset = merge_datasets(list_train_datasets)
        test_dataset = self.sub_datasets[idx_fold]

        return train_dataset, test_dataset


class Metric:
    def __init__(self,
                 mae: float,
                 rmse: float,
                 mape: float,
                 r2: float):
        """
        A data class to store evaluation metrics.

        :param mae: (*float*) Mean absolute error.
        :param rmse: (*float*) Root mean square error.
        :param mape: (*float*) Mean absolute percentage error.
        :param r2: (*float*) The coefficient of determination (R2 score).
        """

        self.mae = mae
        self.rmse = rmse
        self.mape = mape
        self.r2 = r2

    def print(self):
        print('#####################################################')
        print('# Evaluation Metrics of the Prediction Results')
        print('#####################################################')
        print('# Mean Absolute Error (MAE): {:.4f}'.format(self.mae))
        print('# Root Mean Square Error (RMSE): {:.4f}'.format(self.rmse))
        print('# Mean Absolute Percentage Error (MAPE): {:.4f}'.format(self.mape))
        print('# R2 Score: {:.4f}'.format(self.r2))
        print('#####################################################')


def merge_datasets(list_datasets: list):
    """
    Generate a dataset object from a list of datasets in ``list_datasets``.

    :param list_datasets: (*list*) A list of datasets.
    :return: Dataset object.
    """

    data = numpy.vstack([dataset.data for dataset in list_datasets])
    idx_data = numpy.hstack([dataset.idx_data for dataset in list_datasets])

    if isinstance(list_datasets[0], VectorDataset):
        dataset = VectorDataset(data, list_datasets[0].idx_feat, list_datasets[0].idx_target,
                                list_datasets[0].var_names, idx_data)
    elif isinstance(list_datasets[0], FormDataset):
        dataset = FormDataset(data, list_datasets[0].idx_form, list_datasets[0].idx_feat, list_datasets[0].idx_target,
                              list_datasets[0].var_names, idx_data)
    elif isinstance(list_datasets[0], GraphDataset):
        dataset = GraphDataset(data, list_datasets[0].idx_struct, list_datasets[0].idx_feat,
                               list_datasets[0].idx_target, list_datasets[0].var_names, idx_data)
    else:
        raise AssertionError('Unknown dataset type: {}.'.format(type(list_datasets[0])))

    return dataset


def evaluate(targets: numpy.ndarray,
             preds: numpy.ndarray):
    """
    Calculate evaluation metrics for given target values ``targets`` and predicted values ``preds``.
    Four evaluation metrics are calculated:

    - Mean absolute error (MAE).
    - Root mean square error (RMSE).
    - Mean absolute percentage error (MAPE).
    - The coefficient of determination (R2 score).

    Definitions of the evaluation metrics are given in
    `Scikit-learn Regression Metrics <https://mendeleev.readthedocs.io/en/stable/>`_.

    :param targets: (*numpy.ndarray*) Target values of the prediction task.
    :param preds: (*numpy.ndarray*) Predicted values of the machine learning model.
    :return: (*maica.ml.util.Metric*) A data class to store the evaluation metrics.
    """

    mae = mean_absolute_error(targets, preds)
    rmse = numpy.sqrt(mean_squared_error(targets, preds))
    mape = mean_absolute_percentage_error(targets, preds)
    r2 = r2_score(targets, preds)

    return Metric(mae, rmse, mape, r2)


def k_fold_cross_val(dataset: maica.data.base.Dataset,
                     model: maica.ml.base.Model,
                     k: int):
    """
    Train and evaluate machine learning model based on k-fold cross validation.
    After the training and evaluation, means and standard deviations are printed in the standard I/O device.

    :param dataset: (*maica.data.base.Dataset*) A dataset to train machine learning model.
    :param model: (*maica.ml.base.Model*) Machine learning model.
    :param k: (*int*) The number of sub-datasets that will be used for k-fold cross validation.
    """

    k_fold = KFoldGenerator(dataset, k=k)
    metrics = list()

    print('#################################################')
    print('\u0007 Model Training with K-Fold Cross Validation')
    print('#################################################')
    for i in range(0, k):
        # Get train and test dataset form the k-fold generator.
        dataset_train, dataset_test = k_fold.get(idx_fold=i)

        # Copy initial model.
        __model = copy.deepcopy(model)

        # Train the model.
        if __model.alg_name in ALGS_SKLEARN:
            __model.fit(dataset_train.x, dataset_train.y)
            preds_test = __model.predict(dataset_test.x)

        # Evaluate the trained model.
        metric = evaluate(dataset_test.y, preds_test)
        metrics.append(metric)

        print('Fold [{}/{}]\tTest MAE: {:.4f}\tTest R2: {:.4f}'.format(i + 1, k, metric.mae, metric.r2))

    # Collect evaluation metrics.
    list_mae = [m.mae for m in metrics]
    list_rmse = [m.rmse for m in metrics]
    list_mape = [m.mape for m in metrics]
    list_r2 = [m.r2 for m in metrics]

    print('#################################################')
    print('\u0007 Results of K-Fold Cross Validation')
    print('#################################################')
    print('# Average MAE: {:.4f}\u00B1{:.4f}'.format(numpy.mean(list_mae), numpy.std(list_mae)))
    print('# Average RMSE: {:.4f}\u00B1{:.4f}'.format(numpy.mean(list_rmse), numpy.std(list_rmse)))
    print('# Average MAPE: {:.4f}\u00B1{:.4f}'.format(numpy.mean(list_mape), numpy.std(list_mape)))
    print('# Average R2 Score: {:.4f}\u00B1{:.4f}'.format(numpy.mean(list_r2), numpy.std(list_r2)))
    print('#################################################')


def get_model(alg_name: str,
              **kwargs):
    """
    Get a machine learning model for given algorithm name and model hyperparameters.
    The names of the algorithms are defined in ``maica_old.core.env``.

    :param alg_name: (*str*) A name of the machine learning algorithm (defined in ``maica_old.core.env``).
    :param kwargs: (*optional*) A dictionary containing model hyperparameters.
    :return: (*maica.ml.base.Model*) A machine learning model.
    """

    if alg_name in env.ALGS_SKLEARN:
        # Scikit-learn Algorithms
        return SKLearnModel(alg_name, **kwargs)
    elif alg_name in env.ALGS_PYTORCH:
        return None
        # Pytorch Algorithms
        # if alg_name == env.ALG_FCNN:
        #     model = FCNN(dim_in=kwargs['dim_in'], dim_out=kwargs['dim_out'])
        # elif alg_name == env.ALG_ATE:
        #     model = Autoencoder(dim_in=kwargs['dim_in'], dim_latent=kwargs['dim_latent'])
        # elif alg_name == env.ALG_GCN:
        #     readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
        #     n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
        #     model = GCN(n_node_feats=kwargs['n_node_feats'], dim_out=kwargs['dim_out'], n_graphs=n_graphs,
        #                 readout=readout)
        # elif alg_name == env.ALG_GAT:
        #     readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
        #     n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
        #     model = GAT(n_node_feats=kwargs['n_node_feats'], dim_out=kwargs['dim_out'], n_graphs=n_graphs,
        #                 readout=readout)
        # elif alg_name == env.ALG_GIN:
        #     readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
        #     n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
        #     model = GIN(n_node_feats=kwargs['n_node_feats'], dim_out=kwargs['dim_out'], n_graphs=n_graphs,
        #                 readout=readout)
        # elif alg_name == env.ALG_CGCNN:
        #     readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
        #     n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
        #     model = CGCNN(n_node_feats=kwargs['n_node_feats'], n_edge_feats=kwargs['n_edge_feats'],
        #                   dim_out=kwargs['dim_out'], n_graphs=n_graphs, readout=readout)
        # else:
        #     raise AssertionError('Invalid request received with unknown algorithm name: {}.'.format(alg_name))

        # Move model parameters from CPU to GPU when sys.run_gpu is enabled.
        if sys.run_gpu:
            model.gpu()

        return model



"""
Data Utilities
--------------
This module contains useful classes and functions for data pre-processing.
It can be used to convert user datasets into new dataset suitable for machine learning.
"""


import math
import numpy
from sklearn.impute import KNNImputer
from maica_old.core.env import *


def impute(data: numpy.ndarray,
           method: str):
    """
    Fill empty values in the given ``data``.
    It is useful for machine learning with the experimental data containing missing values.
    Three imputation methods are provided:

    - Mean-based imputation method.
    - Zero-based imputation method.
    - KNN-based imputation method.

    Possible values of ``method`` selecting the imputation method are given in ``maica_old.core.env``.

    :param data: (*numpy.ndarray*) Data object.
    :param method: (*str*) A key of the imputation method.
    :return: Data object.
    """

    if method == IMPUTE_MEAN:
        # Fill empty values by mean values.
        means = numpy.nanmean(data, axis=0)
        for i in range(0, data.shape[0]):
            for j in range(0, data.shape[1]):
                if math.isnan(float(data[i, j])):
                    data[i, j] = means[j]
    elif method == IMPUTE_ZERO:
        # Fill empty values by zero.
        for i in range(0, data.shape[0]):
            for j in range(0, data.shape[1]):
                if math.isnan(float(data[i, j])):
                    data[i, j] = 0
    elif method == IMPUTE_KNN:
        # Fill empty values using the values of k-nearest neighbor data.
        imputer = KNNImputer(n_neighbors=3)
        data = imputer.fit_transform(data)

    return data


def get_split_idx(n_data: int,
                  ratio: float):
    """
    Get indices that randomly split a set of {0, 1, ..., ``n_data``}.
    The set is divided into two subsets at a ratio of ``ratio`` to ``1 - ratio``.

    :param n_data: (*int*) The number of data in your data object.
    :param ratio: (*float*) The ratio for division.
    :return: Data indices of the two subsets.
    """

    if ratio >= 1 or ratio <= 0:
        raise AssertionError('Ratio must be in (0, 1), but the given ratio is {:.4f}'.format(ratio))

    n_dataset1 = int(ratio * n_data)
    idx_rand = numpy.random.permutation(n_data)

    return idx_rand[:n_dataset1], idx_rand[n_dataset1:]


def get_one_hot_feat(categories: list,
                     hot_category: object):
    """
    Generate one-hot encoding for the argument ``hot_category`` in the categories of ``categories``.

    :param categories: (*list*) A list of categories for the one-hot encoding.
    :param hot_category: (*object*) An emerged class in the one-hot encoding.
    :return: The one-hot encoding feature.
    """

    one_hot_feat = dict()

    for cat in categories:
        one_hot_feat[cat] = 0

    if hot_category in categories:
        one_hot_feat[hot_category] = 1

    return numpy.array(list(one_hot_feat.values()))



"""
System Utilities
--------------------------
This is a core module of MAICA that includes base utilities of the library.
"""


import torch


# Global configurations of machine learning algorithms.
run_gpu = False


def set_gpu_runnable(runnable: bool):
    """
    Set GPU enable for running machine learning algorithm in GPU.

    :param runnable: A flag to indicate the GPU enable.
    :return: None

    Example:

    >>> set_gpu_runnable(True) # Enable GPU.
    """

    if torch.cuda.is_available():
        global run_gpu
        run_gpu = runnable
    else:
        print('GPU is not available in your system. Make sure the CUDA Driver is installed in the system.')



"""
Crystal Structure
-----------------
This module provides several utilities to handle the datasets of the crystsal structures.
Users can load the dataset of the crystal structures to ``GraphDataset`` object by calling ``load_dataset`` function.
"""


import numpy
from tqdm import tqdm
from maica_old.chem.base import load_elem_feats
from maica_old.data.base import read_data_file
from maica_old.data.util import impute
from maica_old.chem.crystal import even_samples
from maica_old.chem.crystal import get_crystal_graph
from maica_old.data.graph import GraphDataset


def load_dataset(path_dataset: str,
                 path_metadata_file: str,
                 idx_mid: numpy.ndarray,
                 idx_feat: numpy.ndarray,
                 idx_target: int,
                 impute_method: str = 'mean',
                 path_elem_embs: str = None):
    """
    Generate the ``GraphDataset`` object for a dataset of the crystal structures.
    It can be used for machine learning to predict materials properties from the crystal structures.

    :param path_dataset: (*str*) A path of the directory storing the CIF files.
    :param path_metadata_file: (*str*) A path of the metadata file of the dataset.
    :param idx_mid: (*numpy.ndarray*) Indices of the variables indicating the crystal structures in the dataset file.
    :param idx_feat: (*object*) Indices of the input numerical features in the data file.
    :param idx_target: (*object*) An index of the target variable in the data file.
    :param impute_method: (*str, optional*) A imputation method to fill empty data in the data file (default = ``maica_old.core.env.IMPUTE_KNN``).
    :param path_elem_embs: (*str, optional*) A path of JSON file storing user-defined elemental embeddings (*default* = ``None``).
    :return: A ``GraphDataset`` object.
    """

    elem_feats = load_elem_feats(path_elem_embs)
    rbf_means = even_samples(min_val=0, max_val=3, n_samples=64)
    metadata_file, metadata = read_data_file(path_metadata_file)
    dataset = list()
    idx_data = list()

    # Read numerical features.
    if idx_feat is not None:
        idx_feat = numpy.atleast_1d(idx_feat)
        metadata[:, idx_feat] = impute(metadata[:, idx_feat], impute_method)

    # Read crystal structures.
    idx_mid = numpy.atleast_1d(idx_mid)
    for i in tqdm(range(0, metadata.shape[0])):
        numeric_feats = None if idx_feat is None else metadata[i, idx_feat]
        targets = None if idx_target is None else metadata[i, idx_target]
        list_graphs = list()

        # If the data contains multiple crystal structures.
        for j in range(0, idx_mid.shape[0]):
            path_cif = path_dataset + '/' + str(metadata[i, idx_mid[j]]) + '.cif'
            mg = get_crystal_graph(elem_feats, path_cif, numeric_feats, rbf_means, targets, gid=i)

            if mg is not None:
                list_graphs.append(mg)

        if len(list_graphs) == idx_mid.shape[0]:
            dataset.append(list_graphs)
            idx_data.append(i)

    # Generate a graph dataset containing the crystal structures.
    dataset = GraphDataset(dataset, idx_mid, idx_feat, idx_target, metadata_file.columns, numpy.array(idx_data))

    return dataset



"""
Data Utilities
--------------
This module contains useful classes and functions for data pre-processing.
It can be used to convert user datasets into new dataset suitable for machine learning.
"""


import math
import numpy
import pandas
from sklearn.impute import KNNImputer
from rdkit import Chem
from pymatgen.core.structure import Structure
from maica.core.env import *
from maica.core.sys import *
from maica.chem.base import atom_nums
from maica.chem.base import atom_syms
from maica.chem.formula import parse_form


def impute(data: numpy.ndarray,
           method: str):
    """
    Fill empty values in the given ``data``.
    It is useful for machine learning with the experimental data containing missing values.
    Three imputation methods are provided:

    - Mean-based imputation method.
    - Zero-based imputation method.
    - KNN-based imputation method.

    Possible values of ``method`` selecting the imputation method are given in ``maica.core.env``.

    :param data: (*numpy.ndarray*) Data object.
    :param method: (*str*) A key of the imputation method.
    :return: (*numpy.ndarray*) Data object.
    """

    if method == IMPUTE_MEAN:
        # Fill empty values by mean values.
        means = numpy.nanmean(data, axis=0)
        for i in range(0, data.shape[0]):
            for j in range(0, data.shape[1]):
                if math.isnan(float(data[i, j])):
                    data[i, j] = means[j]
    elif method == IMPUTE_ZERO:
        # Fill empty values by zero.
        for i in range(0, data.shape[0]):
            for j in range(0, data.shape[1]):
                if math.isnan(float(data[i, j])):
                    data[i, j] = 0
    elif method == IMPUTE_KNN:
        # Fill empty values using the values of k-nearest neighbor data.
        imputer = KNNImputer(n_neighbors=3)
        data = imputer.fit_transform(data)
    else:
        raise AssertionError(ERR_UNKNOWN_IMPUTE.format(method))

    return data


def get_split_idx(n_data: int,
                  ratio: float):
    """
    Get indices that randomly split a set of {0, 1, ..., ``n_data``}.
    The set is divided into two subsets at a ratio of ``ratio`` to ``1 - ratio``.

    :param n_data: (*int*) The number of data in your data object.
    :param ratio: (*float*) The ratio for division.
    :return: (*numpy.ndarray, numpy.ndarray*) Data indices of the two subsets.
    """

    if ratio >= 1 or ratio <= 0:
        raise AssertionError(ERR_INVALID_SPLIT_RATIO.format(ratio))

    n_dataset1 = int(ratio * n_data)
    idx_rand = numpy.random.permutation(n_data)

    return idx_rand[:n_dataset1], idx_rand[n_dataset1:]


def get_one_hot_feat(categories: list,
                     hot_category: object):
    """
    Generate one-hot encoding for the argument ``hot_category`` in the categories of ``categories``.

    :param categories: (*list*) A list of categories for the one-hot encoding.
    :param hot_category: (*object*) An emerged class in the one-hot encoding.
    :return: (*numpy.ndarray*) The one-hot encoding feature.
    """

    one_hot_feat = dict()

    for cat in categories:
        one_hot_feat[cat] = 0

    if hot_category in categories:
        one_hot_feat[hot_category] = 1

    return numpy.array(list(one_hot_feat.values()))


def get_elem_dist(path_metadata_file: str,
                  idx_chem_code: int,
                  data_type: str,
                  path_cif_files: str = None):
    """
    Calculate elemental distribution in a dataset.
    For given metadata in ``path_metadata_file`` and data type in ``data_type``,
    it calculate the frequency for each element in the chemical compounds in the dataset.
    A dictionary object containing the frequency for each element is returned.

    :param path_metadata_file: (*str*) A path of the metadata file of the dataset.
    :param idx_chem_code: (*int*) An index of the chemical compound.
    :param data_type: (*str*) A data type of the chemical compound, which is available in {``DATA_TYPE_FORM``, ``DATA_TYPE_SMILES``, ``DATA_TYPE_CIF``}.
    :param path_cif_files: (*str, optional*) A path of the cif files for a data type of ``DATA_TYPE_CIF``.
    :return: (*dict*) The frequency for each element.
    """

    metadata = numpy.array(pandas.read_excel(path_metadata_file))
    elems = atom_nums.keys()
    elem_dist = dict()

    # Initialize a dictionary storing the elemental distribution.
    for e in elems:
        elem_dist[e] = 0

    if data_type == DATA_TYPE_FORM:
        # For chemical formulas.
        for i in range(0, metadata.shape[0]):
            form = parse_form(metadata[i, idx_chem_code])
            for e in form.keys():
                elem_dist[e] += 1
    elif data_type == DATA_TYPE_SMILES:
        # For molecular structures in SMILES.
        for i in range(0, metadata.shape[0]):
            mol = Chem.MolFromSmiles(metadata[i, idx_chem_code])
            atom_syms_mol = numpy.unique([atom_syms[atom.GetAtomicNum()] for atom in mol.GetAtoms()])
            for e in atom_syms_mol:
                elem_dist[e] += 1
    elif data_type == DATA_TYPE_CIF:
        # For crystal structures in CIF files.
        if path_cif_files is None:
            raise AssertionError('Path of the CIF files must be provided by path_cif_file'
                                 ' for a data type of DATA_TYPE_CIF.')

        for i in range(0, metadata.shape[0]):
            crystal = Structure.from_file(path_cif_files + '/' + metadata[i, idx_chem_code] + '.cif')
            atom_syms_crystal = numpy.unique([atom_syms[atom_num] for atom_num in crystal.atomic_numbers])
            for e in atom_syms_crystal:
                elem_dist[e] += 1
    else:
        raise AssertionError('Unknown data type {} was given'.format(data_type))

    return elem_dist



"""
Machine Learning Utilities
--------------------------
The ``maica_old.ml.util`` module provides essential functions for training configuration and model reuse.
Most deep learning algorithms in MAICA are based on this module.
"""

import numpy
import torch
import pandas
import os
import graphviz
import xgboost
import maica_old.ml.base
import maica_old.data.base
import torch.utils.data as tdata
import torch_geometric.data as tgdata
import maica_old.core.sys as sys
import maica_old.core.env as env
from sklearn.tree import export_graphviz
from maica_old.ml.base import SKLearnModel
from maica_old.ml.nn import FCNN
from maica_old.ml.nn import Autoencoder
from maica_old.ml.gnn import GCN
from maica_old.ml.gnn import GAT
from maica_old.ml.gnn import GIN
from maica_old.ml.gnn import CGCNN
from maica_old.data.base import Dataset
from maica_old.data.vector import VectorDataset
from maica_old.data.formula import FormDataset
from maica_old.data.graph import GraphDataset


class KFoldGenerator:
    def __init__(self,
                 dataset: Dataset,
                 k: int):
        self.sub_datasets = dataset.get_sub_datasets(k=k)

    def get(self,
            idx_fold: int):
        """
        Get training and test datasets for a given ``idx_fold``.

        :param idx_fold: (*int*) An index of the k-fold dataset.
        :return: Training and test datasets.
        """

        list_train_datasets = [self.sub_datasets[i] for i in range(0, len(self.sub_datasets)) if i != idx_fold]
        train_dataset = merge_datasets(list_train_datasets)
        test_dataset = self.sub_datasets[idx_fold]

        return train_dataset, test_dataset


def get_batch_sizes(n_data: int,
                    train_setting: int):
    """
    Return a batch size for stochastic gradient descent method according to the number of data and the train setting.

    :param n_data: (*int*) The number of data will be used to train.
    :param train_setting: (*int*) A hyperparameter determining the train settings.
    :return: A list of candidate batch sizes.
    """

    if n_data < 16:
        return [n_data]
    elif n_data < 32:
        return [16]
    elif n_data < 128:
        return [32]
    elif n_data < 256:
        return [64]
    else:
        # For data with a size >= 256.
        if train_setting == 0:
            return [64]
        elif train_setting == 1:
            return [64, 128]
        elif train_setting == 2:
            return [32, 64]
        elif train_setting == 3:
            return [32, 64, 128]
        elif train_setting == 4:
            return [32, 64, 128, 256]


def get_init_lrs(hparam_setting: int):
    """
    Return initial learning rates according to the given hyperparameter setting level.

    :param hparam_setting: (*int*) A hyperparameter determining the hyperparameter optimization of the model.
    :return: A list of candidate learning rates.
    """

    if hparam_setting == 0:
        return [1e-3]
    elif hparam_setting == 1:
        return [5e-3, 1e-3]
    elif hparam_setting == 2:
        return [5e-3, 1e-3, 5e-4]
    elif hparam_setting == 3:
        return [1e-2, 5e-3, 1e-3, 5e-4]
    elif hparam_setting == 4:
        return [5e-3, 1e-2, 5e-3, 1e-3, 5e-4, 1e-4]


def get_data_loader(*data: object,
                    batch_size: int = 8,
                    shuffle: bool = False):
    """
    Generate data loader object for a given dataset.
    If the given data is ``numpy.ndarray``, it returns ``torch.DataLoader`` object.
    If the data is ``maica_old.data.GraphDataset``, it returns ``torch_geometric.DataLoader`` object to iterate the graph-structured data.

    :param data: (*object*) The dataset to be iterated by the data loader.
    :param batch_size: (*int, optional*) The batch size of the data loader (*default* = 8).
    :param shuffle: (*int, optional*) An option to randomly sample the data when the iterations of the data loader (*default* = ``False``).
    :return: Data loader object of ``torch.DataLoader`` or ``torch_geometric.DataLoader``.
    """

    if isinstance(data[0], numpy.ndarray):
        # Generate data loader for the vector dataset including the numerical vectors and the chemical formulas.
        tensors = [torch.tensor(d, dtype=torch.float) for d in data]
        return tdata.DataLoader(tdata.TensorDataset(*tuple(tensors)), batch_size=batch_size, shuffle=shuffle)
    elif isinstance(data[0], GraphDataset):
        # Generate data loader for the graph dataset including the numerical vectors and the chemical formulas.
        return tgdata.DataLoader(data[0].data, batch_size=batch_size, shuffle=shuffle)
    else:
        raise AssertionError('The type of the given data object(s) is not valid.' +
                             ' Only numpy.ndarray and ml.data.Dataset is acceptable for this function.')


def get_optimizer(model_params: torch.Generator,
                  gd_name: str,
                  init_lr: float = 1e-3,
                  l2_reg: float = 1e-6):
    """
    Return a gradient descent optimizer to fit model parameters.

    :param model_params: (*torch.Generator*) Model parameters to be trained by the generated optimizer.
    :param gd_name: (*str*) A name of the gradient descent method to fit model parameters (defined in ``maica_old.core.env``).
    :param init_lr: (*float, optional*) An initial learning rate of the gradient descent optimizer.
    :param l2_reg: (*float, optional*) A coefficient of the L2 regularization in model parameters.
    :return: A gradient descent optimizer to fit model parameters.
    """

    # Define the gradient descent method to optimize the model parameters.
    if gd_name == env.GD_SGD:
        optimizer = torch.optim.SGD(model_params, lr=init_lr, weight_decay=l2_reg, momentum=0.9)
    elif gd_name == env.GD_ADADELTA:
        optimizer = torch.optim.Adadelta(model_params, lr=init_lr, weight_decay=l2_reg)
    elif gd_name == env.GD_RMSPROP:
        optimizer = torch.optim.RMSprop(model_params, lr=init_lr, weight_decay=l2_reg)
    elif gd_name == env.GD_ADAM:
        optimizer = torch.optim.Adam(model_params, lr=init_lr, weight_decay=l2_reg)
    else:
        raise AssertionError('Invalid request received with unknown name of the gradient method: {}.'.format(gd_name))

    return optimizer


def get_loss_func(loss_func: str):
    """
    Return a loss function to evaluate the model performance.

    :param loss_func: (*str*) A name of the loss function to evaluate model performance (defined in `maica_old.core.env``).
    :return: A loss function object to evaluate the model performance.
    """

    # Define the loss function to evaluate the model performance.
    if loss_func == env.LOSS_MAE:
        criterion = torch.nn.L1Loss()
    elif loss_func == env.LOSS_MSE:
        criterion = torch.nn.MSELoss()
    elif loss_func == env.LOSS_SMAE:
        criterion = torch.nn.SmoothL1Loss()
    else:
        raise AssertionError('Invalid request received with unknown name of the loss function: {}'.format(loss_func))

    return criterion


def get_model(alg_name: str,
              **kwargs):
    """
    Get a machine learning model for given algorithm name and model hyperparameters.
    The names of the algorithms are defined in ``maica_old.core.env``.

    :param alg_name: (*str*) A name of the machine learning algorithm (defined in ``maica_old.core.env``).
    :param kwargs: (*optional*) A dictionary containing model hyperparameters.
    :return: A machine learning model.
    """

    if alg_name in env.ALGS_SKLEARN:
        # Scikit-learn Algorithms
        return SKLearnModel(alg_name, **kwargs)
    elif alg_name in env.ALGS_PYTORCH:
        # Pytorch Algorithms
        if alg_name == env.ALG_FCNN:
            model = FCNN(dim_in=kwargs['dim_in'], dim_out=kwargs['dim_out'])
        elif alg_name == env.ALG_ATE:
            model = Autoencoder(dim_in=kwargs['dim_in'], dim_latent=kwargs['dim_latent'])
        elif alg_name == env.ALG_GCN:
            readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
            n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
            model = GCN(n_node_feats=kwargs['n_node_feats'], dim_out=kwargs['dim_out'], n_graphs=n_graphs,
                        readout=readout)
        elif alg_name == env.ALG_GAT:
            readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
            n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
            model = GAT(n_node_feats=kwargs['n_node_feats'], dim_out=kwargs['dim_out'], n_graphs=n_graphs,
                        readout=readout)
        elif alg_name == env.ALG_GIN:
            readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
            n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
            model = GIN(n_node_feats=kwargs['n_node_feats'], dim_out=kwargs['dim_out'], n_graphs=n_graphs,
                        readout=readout)
        elif alg_name == env.ALG_CGCNN:
            readout = kwargs['readout'] if 'readout' in kwargs.keys() else 'mean'
            n_graphs = kwargs['n_graphs'] if 'n_graphs' in kwargs.keys() else 1
            model = CGCNN(n_node_feats=kwargs['n_node_feats'], n_edge_feats=kwargs['n_edge_feats'],
                          dim_out=kwargs['dim_out'], n_graphs=n_graphs, readout=readout)
        else:
            raise AssertionError('Invalid request received with unknown algorithm name: {}.'.format(alg_name))

        # Move model parameters from CPU to GPU when sys.run_gpu is enabled.
        if sys.run_gpu:
            model.gpu()

        return model


def save_eval_results(task_name: str,
                      model: maica_old.ml.base.Model,
                      dataset_test: maica_old.data.base.Dataset,
                      preds: numpy.ndarray):
    """
    Save the model parameters and the prediction results as a model file and an excel file.

    :param task_name: (*str*) A name of your task.
    :param model: (*ml.base.Model)* A model used will be evaluated.
    :param dataset_test: (*data.base.Dataset*) A dataset used to the evaluation.
    :param preds: (*numpy.ndarray*) Prediction results of the model for the dataset.
    """

    # Make a directory to save the evaluation results.
    if not os.path.exists(task_name):
        os.mkdir(task_name)

    # Save the trained model.
    if model.alg_type == env.ALG_SKLEARN:
        model.save(task_name + '/model_' + model.alg_name + '.joblib')
    elif model.alg_type == env.ALG_PYTORCH:
        model.save(task_name + '/model_' + model.alg_name + '.pt')
    else:
        raise AssertionError('Unknown algorithm type: {}.'.format(model.alg_type))

    # Save the prediction results.
    idx_data = dataset_test.idx_data.reshape(-1, 1)
    targets = dataset_test.y.reshape(-1, 1)
    _preds = preds.reshape(-1, 1)
    df = pandas.DataFrame(numpy.hstack([idx_data, targets, _preds]))
    df.to_excel(task_name + '/pred_results_' + model.alg_name + '.xlsx',
                index=None, header=['data_index', 'target', 'prediction'])


def save_interpretation(model: maica_old.ml.base.Model,
                        file_name: str):
    """
    Save interpretable information of the machine learning algorithms.
    For non-interpretable algorithms, it raises ``AssertionError``.
    The following algorithms support this function.

    - Decision Tree Regression (ALG_DCTR).
    - Symbolic Regression (ALG_SYMR).
    - Gradient Boosting Tree Regression (ALG_GBTR).

    :param model: (*ml.base.Model*) A machine learning algorithm to generate interpretation about the prediction.
    :param file_name: (*str*) The path of file to store the generated interpretation.
    """

    if model.alg_name == env.ALG_DCTR:
        struct = export_graphviz(model.alg, feature_names=['feat #' + str(i) for i in range(0, model.alg.n_features_)])
        graphviz.Source(struct, format='png').render(file_name.replace('.png', ''), format='png')
    elif model.alg_name == env.ALG_SYMR:
        expression = model.alg._program.export_graphviz()
        graphviz.Source(expression, format='png').render(file_name.replace('.png', ''), format='png')
    elif model.alg_name == env.ALG_GBTR:
        xgboost.plot_importance(model.alg).figure.savefig(file_name)
    else:
        raise AssertionError('Interpretation is available for only the following algorithms.' +
                             '\n1. Decision Tree Regression.' +
                             '\n2. Symbolic Regression.' +
                             '\n3. Gradient Boosting Tree Regression.')


def merge_datasets(list_datasets: list):
    """
    Generate a dataset object from a list of datasets in ``list_datasets``.

    :param list_datasets: (*list*) A list of datasets.
    :return: Dataset object.
    """

    data = numpy.vstack([dataset.data for dataset in list_datasets])
    idx_data = numpy.hstack([dataset.idx_data for dataset in list_datasets])

    if isinstance(list_datasets[0], VectorDataset):
        dataset = VectorDataset(data, list_datasets[0].idx_feat, list_datasets[0].idx_target,
                                list_datasets[0].var_names, idx_data)
    elif isinstance(list_datasets[0], FormDataset):
        dataset = FormDataset(data, list_datasets[0].idx_form, list_datasets[0].idx_feat, list_datasets[0].idx_target,
                              list_datasets[0].var_names, idx_data)
    elif isinstance(list_datasets[0], GraphDataset):
        dataset = GraphDataset(data, list_datasets[0].idx_struct, list_datasets[0].idx_feat,
                               list_datasets[0].idx_target, list_datasets[0].var_names, idx_data)
    else:
        raise AssertionError('Unknown dataset type: {}.'.format(type(list_datasets[0])))

    return dataset



"""
System Utilities
----------------
It is a core module of MAICA to provide base utilities of the framework.
"""


import torch


# maica.data.base.read_data_file
ERR_DATA_FILE_EXT = 'Only .xlsx and .csv extensions are available, but unknown file extension {} was given.'

# maica.data.util.impute
ERR_UNKNOWN_IMPUTE = 'Unknown imputation method {}.'

# maica.data.util.get_split_idx
ERR_INVALID_SPLIT_RATIO = 'Ratio must be in (0, 1), but the given ratio is {:.4f}'

# maica.data.vector.Vector.denormalize
ERR_UNNORMALIZED = 'The input data of this dataset has never been normalized.'

# maica.data.vector.load_dataset
ERR_EMPTY_FEAT_IDX = 'Indices of the input features must be provided to load the dataset object.'

# maica.data.formula.load_dataset
ERR_EMPTY_FORM_IDX = 'Indices of the chemical formulas must be provided to load the dataset object.'

# maica.data.molecule.load_dataset, maica.data.crystal.load_dataset
ERR_EMPTY_STRUCT_IDX = 'Indices of the structures must be provided to load the dataset object.'

# maica.util.optimization.run_ml_model
ERR_UNKNOWN_ALG = 'The type of the algorithm {} is unknown.'

# maica.ml.base.Model
ERR_UNKNOWN_PIPELINE = 'Unknown pipeline type {} was given.'


# Global configurations of machine learning algorithms.
run_gpu = False


def set_gpu_runnable(runnable: bool):
    """
    Set GPU enable for running machine learning algorithm in GPU.

    :param runnable: A flag to indicate the GPU enable.
    :return: None

    Example:

    >>> set_gpu_runnable(True) # Enable GPU.
    """

    if torch.cuda.is_available():
        global run_gpu
        run_gpu = runnable
    else:
        print('GPU is not available in your system. Make sure the CUDA Driver is installed in the system.')



"""
Chemical Formula
----------------
This module includes a basic class ``FormDataset`` to load a dataset containing the chemical formulas.
"""


import numpy
import copy
import maica_old.core.env
from tqdm import tqdm
from maica_old.data.vector import VectorDataset
from maica_old.data.base import read_data_file
from maica_old.data.util import get_split_idx
from maica_old.data.util import impute
from maica_old.chem.base import load_elem_feats
from maica_old.chem.formula import form_to_vec
from maica_old.chem.formula import form_to_sparse_vec
from maica_old.chem.formula import parse_form


class FormDataset(VectorDataset):
    """
    It is a dataset class to load datasets based on the chemical formulas.
    To generate this class from your dataset, call ``load_dataset`` with a path of your dataset.
    """

    def __init__(self,
                 dataset: numpy.ndarray,
                 idx_form: numpy.ndarray,
                 idx_feat: numpy.ndarray,
                 idx_target: int,
                 var_names: numpy.ndarray,
                 forms: numpy.ndarray,
                 idx_data: numpy.ndarray = None):
        """
        :param dataset: (*numpy.ndarray*) A NumPy array of the data object.
        :param idx_form: (*numpy.ndarray*) Indices of the chemical formulas in the dataset file.
        :param idx_feat: (*numpy.ndarray*) Indices of the numerical features in the dataset file.
        :param idx_target: (*int*) An index of the target value in the dataset file.
        :param var_names: (*numpy.ndarray*) Names of all variables in the dataset file.
        :param forms: (*numpy.ndarray*) The chemical formulas of each data in the dataset.
        :param idx_data: (*numpy.ndarray*) Numerical indices of the data (*default* = ``None``).
        """

        super(FormDataset, self).__init__(dataset, idx_feat, idx_target, var_names, idx_data)
        self.idx_form = copy.deepcopy(idx_form)
        self.forms = copy.deepcopy(forms)

        if type(self).__name__ == 'FormDataset':
            self._set_feat_names()
            self._set_tooltips()

    def _set_feat_names(self):
        """
        Set names and types of the input features.
        """

        for idx in self.idx_form:
            self.feat_names.append(self.var_names[idx])
            self.feat_types.append('form')

        if self.idx_feat is not None:
            for idx in self.idx_feat:
                self.feat_names.append(self.var_names[idx])
                self.feat_types.append('num')

    def _set_tooltips(self):
        """
        Set tooltip for each data to identify it in the data visualization.
        """

        for i in range(0, self.data.shape[0]):
            self.tooltips.append('Data idx: ' + str(self.idx_data[i]) + ' (' + ' '.join(self.forms[i]) + ')')

    def split(self,
              ratio: float):
        """
        Split a dataset into two sub-datasets based on the given ratio.
        Two sub-datasets and the original indices of the data in them are returned.

        :param ratio: (*float*) Ratio between two sub-datasets. The sub-datasets are dived by a ratio of ``ratio`` to ``1 - ratio``.
        :return: Two sub-datasets and the original indices of the data in them.
        """

        # Randomly sample the dataset.
        idx_dataset1, idx_dataset2 = get_split_idx(self.data.shape[0], ratio)

        # Split dataset into the two subsets.
        dataset1 = FormDataset(self.data[idx_dataset1, :], self.idx_form, self.idx_feat, self.idx_target,
                               self.var_names, self.forms[idx_dataset1], self.idx_data[idx_dataset1])
        dataset2 = FormDataset(self.data[idx_dataset2, :], self.idx_form, self.idx_feat, self.idx_target,
                               self.var_names, self.forms[idx_dataset2], self.idx_data[idx_dataset2])

        return dataset1, dataset2

    def get_sub_datasets(self,
                         k: int):
        """
        Split the dataset into the :math:`k` sub-datasets without repeating the data.
        In the training, :math:`k-1` sub-datasets are used to train the model, and the remaining sub-dataset is used for evaluation.

        :param k: (*int*) The number of subsets.
        :return: A list of :math:`k` sub-datasets.
        """

        idx_rand = numpy.random.permutation(self.n_data)
        n_data_subset = int(self.n_data / k)
        sub_datasets = list()

        # Get k-1 sub-datasets with the same size.
        for i in range(0, k-1):
            idx_sub_dataset = idx_rand[i*n_data_subset:(i+1)*n_data_subset]
            sub_dataset = FormDataset(self.data[idx_sub_dataset, :], self.idx_form, self.idx_feat, self.idx_target,
                                      self.var_names, self.forms[idx_sub_dataset], self.idx_data[idx_sub_dataset])
            sub_datasets.append(sub_dataset)

        # Get the last sub-dataset containing the all remaining data.
        idx_sub_dataset = idx_rand[(k-1)*n_data_subset:]
        sub_dataset = FormDataset(self.data[idx_sub_dataset, :], self.idx_form, self.idx_feat, self.idx_target,
                                  self.var_names, self.forms[idx_sub_dataset], self.idx_data[idx_sub_dataset])
        sub_datasets.append(sub_dataset)

        return sub_datasets


def load_dataset(path_data_file: str,
                 idx_form: object,
                 idx_feat: object,
                 idx_target: int,
                 impute_method: str = maica_old.core.env.IMPUTE_KNN,
                 path_elem_embs: str = None,
                 rep_type: str = 'compact'):
    """
    This is a wrapper function to load a dataset containing the chemical formulas.
    It loads a dataset object from a given dataset file in ``maica_old.data.vector.VectorDataset``.
    A representation method for the chemical formulas can be selected according to a given ``rep_type``.

    :param path_data_file: (*str*) The path of the data file.
    :param idx_form: (*object*) Indices of the chemical formulas in the data file.
    :param idx_feat: (*object*) Indices of the input numerical features in the data file.
    :param idx_target: (*int*) An index of the target variable in the data file.
    :param impute_method: (*str, optional*) A imputation method to fill empty data in the data file (*default* = ``maica_old.core.env.IMPUTE_KNN``).
    :param path_elem_embs: (*str, optional*) A path of JSON file storing user-defined elemental embeddings (*default* = ``None``).
    :param rep_type: (*str, optional*) A representation way to describe the chemical formulas (*default* = ``compact``).
    :return: A ``FormDataset`` object.
    """

    if rep_type == 'compact':
        return __load_dataset_compact(path_data_file, idx_form, idx_feat, idx_target, impute_method, path_elem_embs)
    elif rep_type == 'sparse':
        return __load_dataset_sparse(path_data_file, idx_form, idx_feat, idx_target, impute_method, path_elem_embs)


def __load_dataset_compact(path_data_file: str,
                           idx_form: object,
                           idx_feat: object,
                           idx_target: int,
                           impute_method: str,
                           path_elem_embs: str):
    """
    Load a dataset object from a dataset file in ``path_data_file`` containing the chemical formulas.
    All chemical formulas are represented as feature vectors with the same dimension based on the statistics of atomic features.
    A details description of the compact feature vector is given in the API reference of ``maica_old.chem.formula.form_to_vec``.

    :param path_data_file: (*str*) The path of the data file.
    :param idx_form: (*object*) Indices of the chemical formulas in the data file.
    :param idx_feat: (*object*) Indices of the input numerical features in the data file.
    :param idx_target: (*int*) An index of the target variable in the data file.
    :param impute_method: (*str, optional*) A imputation method to fill empty data in the data file (*default* = ``maica_old.core.env.IMPUTE_KNN``).
    :param path_elem_embs: (*str, optional*) A path of JSON file storing user-defined elemental embeddings (*default* = ``None``).
    :return: A ``FormDataset`` object.
    """

    idx_form = numpy.atleast_1d(idx_form)
    elem_feats = load_elem_feats(path_elem_embs)
    data_file, data = read_data_file(path_data_file)
    var_names = data_file.columns.values
    dataset = list()

    # Convert chemical formulas into numeric vectors based on the elemental embeddings.
    # Iterate idx_form and data.shape[0] to convert multiple formulas in a data row.
    form_vecs = list()
    for i in tqdm(range(0, data.shape[0])):
        form_feats = list()

        for idx in idx_form:
            form_feats.append(form_to_vec(str(parse_form(data[i, idx])), elem_feats))
        form_vecs.append(numpy.hstack(form_feats))
    dataset.append(numpy.vstack(form_vecs))

    # Check existence of numeric features.
    if idx_feat is not None:
        idx_feat = numpy.atleast_1d(idx_feat)
        dataset.append(impute(data[:, idx_feat], impute_method))

    # Check existence of target property.
    if idx_target is not None:
        dataset.append(data[:, idx_target].reshape(-1, 1))

    # Create FormDataset object.
    dataset = numpy.hstack(dataset).astype(numpy.float)
    form_dataset = FormDataset(dataset, idx_form, idx_feat, idx_target, var_names, data[:, idx_form])

    return form_dataset


def __load_dataset_sparse(path_data_file: str,
                          idx_form: object,
                          idx_feat: object,
                          idx_target: int,
                          impute_method: str,
                          path_elem_embs: str):
    """
    Load a dataset object from a dataset file in ``path_data_file`` containing the chemical formulas.
    The chemical formulas are represented as feature vectors with different dimensions based on the number elements in them.
    A details description of the compact feature vector is given in the API reference of ``maica_old.chem.formula.form_to_sparse_vec``.

    :param path_data_file: (*str*) The path of the data file.
    :param idx_form: (*object*) Indices of the chemical formulas in the data file.
    :param idx_feat: (*object*) Indices of the input numerical features in the data file.
    :param idx_target: (*int*) An index of the target variable in the data file.
    :param impute_method: (*str, optional*) A imputation method to fill empty data in the data file (*default* = ``maica_old.core.env.IMPUTE_KNN``).
    :param path_elem_embs: (*str, optional*) A path of JSON file storing user-defined elemental embeddings (*default* = ``None``).
    :return: A ``FormDataset`` object.
    """

    idx_form = numpy.atleast_1d(idx_form)
    elem_feats = load_elem_feats(path_elem_embs)
    data_file, data = read_data_file(path_data_file)
    var_names = data_file.columns.values
    formulas = list()
    dataset = list()

    # Find maximum number of elements in the chemical formulas.
    for i in range(0, data.shape[0]):
        formulas.append(list())

        for idx in idx_form:
            formulas[i].append(parse_form(data[i, idx]))
    max_elems = numpy.max([len(x.keys()) for d in formulas for x in d])

    form_vecs = list()
    for i in tqdm(range(0, data.shape[0])):
        form_feats = list()

        for formula in formulas[i]:
            form_feats.append(form_to_sparse_vec(str(formula), elem_feats, max_elems))
        form_vecs.append(numpy.hstack(form_feats))
    dataset.append(numpy.vstack(form_vecs))

    # Check existence of numeric features.
    if idx_feat is not None:
        idx_feat = numpy.atleast_1d(idx_feat)
        dataset.append(impute(data[:, idx_feat], impute_method))

    # Check existence of target property.
    if idx_target is not None:
        dataset.append(data[:, idx_target].reshape(-1, 1))

    # Create FormDataset object.
    dataset = numpy.hstack(dataset).astype(numpy.float)
    form_dataset = FormDataset(dataset, idx_form, idx_feat, idx_target, var_names, data[:, idx_form])

    return form_dataset



"""
Graph Neural Networks
---------------------
The ``maica_old.ml.gnn`` module includes various implementation of graph neural networks from the torch_geometric library.
It provides pre-defined graph neural networks for the structure-based predictions.
"""

import numpy
import torch
import torch_geometric
import torch.nn as nn
import torch.nn.functional as F
import maica_old.core.sys as sys
import maica_old.core.env as env
from torch_geometric.nn import LayerNorm
from torch_geometric.nn import GCNConv
from torch_geometric.nn import GATConv
from torch_geometric.nn import GINConv
from torch_geometric.nn import CGConv
from torch_geometric.nn import global_mean_pool
from torch_geometric.nn import global_add_pool
from maica_old.ml.base import PyTorchModel


class GNN(PyTorchModel):
    """
        Abstract class of the graph neural networks.
        It defines two functions for parameter optimization and prediction.
    """

    def __init__(self,
                 dim_out: int,
                 alg_name: str):
        super(GNN, self).__init__(alg_name)
        self.dim_out = dim_out

    def fit(self,
            data_loader: torch_geometric.data.DataLoader,
            optimizer: torch.optim.Optimizer,
            criterion: object):
        """
        Fit the model parameters for the given dataset in the data loader, optimizer, and loss function.
        It iterates the parameter optimization once for the entire dataset.

        :param data_loader: (torch_geometric.data.DataLoader) A data loader to sample the data from the training dataset.
        :param optimizer: (torch.optim.Optimizer) An optimizer to fit the model parameters.
        :param criterion: (object) A loss function to evaluate the prediction performance of the model.
        :return: Training loss.
        """

        self.train()
        sum_losses = 0

        for batch in data_loader:
            if sys.run_gpu:
                for b in batch:
                    b.x = b.x.cuda()
                    b.y = b.y.cuda()
                    b.edge_index = b.edge_index.cuda()
                    b.batch = b.batch.cuda()
                    b.edge_attr = None if b.edge_attr is None else b.edge_attr.cuda()

            preds = self(batch)
            loss = criterion(batch[0].y, preds)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            sum_losses += loss.detach().item()

        return sum_losses / len(data_loader)

    def predict(self,
                data_loader: torch_geometric.data.DataLoader):
        """
        Predict target values for the given dataset in the data loader.

        :param data_loader: (torch_geometric.data.DataLoader) A data loader to sample the data from the dataset.
        :return: A NumPy array containing the predicted values.
        """

        self.eval()
        list_preds = list()

        with torch.no_grad():
            for batch in data_loader:
                if sys.run_gpu:
                    for b in batch:
                        b.x = b.x.cuda()
                        b.y = b.y.cuda()
                        b.edge_index = b.edge_index.cuda()
                        b.batch = b.batch.cuda()
                        b.edge_attr = None if b.edge_attr is None else b.edge_attr.cuda()

                if sys.run_gpu:
                    list_preds.append(self(batch).cpu().numpy())
                else:
                    list_preds.append(self(batch).numpy())

        return numpy.vstack(list_preds).flatten()


# Thomas N. Kipf and Max Welling
# Semi-Supervised Classification with Graph Convolutional Networks
# International Conference on Learning Representations (ICLR) 2017
class GCN(GNN):
    """
    Graph convolutional network (GCN) form the `"Semi-supervised Classification with Graph Convolutional Networks"
    <https://arxiv.org/abs/1609.02907>`_ paper.
    """

    def __init__(self,
                 n_node_feats: int,
                 dim_out: int,
                 n_graphs: int = 1,
                 readout: str = env.READOUT_MEAN):
        super(GCN, self).__init__(dim_out, env.ALG_GCN)
        self.n_graphs = n_graphs
        self.readout = readout
        self.gc1 = GCNConv(n_node_feats, 256)
        self.gn1 = LayerNorm(256)
        self.gc2 = GCNConv(256, 256)
        self.gn2 = LayerNorm(256)
        self.gc3 = GCNConv(256, 256)
        self.gn3 = LayerNorm(256)
        self.fc1 = nn.Linear(self.n_graphs * 256, 32)
        self.fc2 = nn.Linear(32, dim_out)

    def __emb_nodes(self,
                    g: torch_geometric.data.Batch):
        """
        Generate node embeddings using the graph convolutional layers.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Node embeddings.
        """

        h = F.relu(self.gn1(self.gc1(g.x, g.edge_index)))
        h = F.relu(self.gn2(self.gc2(h, g.edge_index)))
        h = F.relu(self.gn3(self.gc3(h, g.edge_index)))

        return h

    def forward(self,
                g: torch_geometric.data.Batch):
        """
        Predict target values for the given Batch object.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Target values.
        """

        if self.readout == env.READOUT_MEAN:
            hg = torch.cat([global_mean_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)
        elif self.readout == env.READOUT_SUM:
            hg = torch.cat([global_add_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)

        hg = F.relu(self.fc1(hg))
        out = self.fc2(hg)

        return out


# Petar Velickovic, Guillem Cucurull, Arantxa Casanova, Adriana Romero, Pietro Lio, and Yoshua Bengio,
# Graph Attention Networks,
# International Conference on Learning Representations (ICLR) 2018
class GAT(GNN):
    """
    Graph attention network (GAT) form the `"Graph Attention Networks"
    <https://arxiv.org/abs/1710.10903>`_ paper.
    """

    def __init__(self,
                 n_node_feats: int,
                 dim_out: int,
                 n_graphs: int = 1,
                 readout: str = env.READOUT_MEAN):
        super(GAT, self).__init__(dim_out, env.ALG_GAT)
        self.n_graphs = n_graphs
        self.readout = readout
        self.gc1 = GATConv(n_node_feats, 256)
        self.gn1 = LayerNorm(256)
        self.gc2 = GATConv(256, 256)
        self.gn2 = LayerNorm(256)
        self.gc3 = GATConv(256, 256)
        self.gn3 = LayerNorm(256)
        self.fc1 = nn.Linear(self.n_graphs * 256, 32)
        self.fc2 = nn.Linear(32, dim_out)

    def __emb_nodes(self,
                    g: torch_geometric.data.Batch):
        """
        Generate node embeddings using the graph convolutional layers.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Node embeddings.
        """

        h = F.relu(self.gn1(self.gc1(g.x, g.edge_index)))
        h = F.relu(self.gn2(self.gc2(h, g.edge_index)))
        h = F.relu(self.gn3(self.gc3(h, g.edge_index)))

        return h

    def forward(self,
                g: torch_geometric.data.Batch):
        """
        Predict target values for the given Batch object.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Target values.
        """

        if self.readout == env.READOUT_MEAN:
            hg = torch.cat([global_mean_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)
        elif self.readout == env.READOUT_SUM:
            hg = torch.cat([global_add_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)

        hg = F.relu(self.fc1(hg))
        out = self.fc2(hg)

        return out


# Keyulu Xu, Weihua Hu, Jure Leskovec, and Stefanie Jegelka,
# How Powerful are Graph Neural Networks?,
# International Conference on Learning Representations (ICLR) 2019
class GIN(GNN):
    """
    Graph isomorphism network (GIN) form the `"How Powerful are Graph Neural Networks?"
    <https://arxiv.org/abs/1810.00826>`_ paper.
    """

    def __init__(self,
                 n_node_feats: int,
                 dim_out: int,
                 n_graphs: int = 1,
                 readout: str = env.READOUT_MEAN):
        super(GIN, self).__init__(dim_out, env.ALG_GIN)
        self.n_graphs = n_graphs
        self.readout = readout
        self.gc1 = GINConv(nn.Linear(n_node_feats, 256))
        self.gn1 = LayerNorm(256)
        self.gc2 = GINConv(nn.Linear(256, 256))
        self.gn2 = LayerNorm(256)
        self.gc3 = GINConv(nn.Linear(256, 256))
        self.gn3 = LayerNorm(256)
        self.fc1 = nn.Linear(self.n_graphs * 256, 32)
        self.fc2 = nn.Linear(32, dim_out)

    def __emb_nodes(self,
                    g: torch_geometric.data.Batch):
        """
        Generate node embeddings using the graph convolutional layers.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Node embeddings.
        """
        h = F.relu(self.gn1(self.gc1(g.x, g.edge_index)))
        h = F.relu(self.gn2(self.gc2(h, g.edge_index)))
        h = F.relu(self.gn3(self.gc3(h, g.edge_index)))

        return h

    def forward(self,
                g: torch_geometric.data.Batch):
        """
        Predict target values for the given Batch object.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Target values.
        """

        if self.readout == env.READOUT_MEAN:
            hg = torch.cat([global_mean_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)
        elif self.readout == env.READOUT_SUM:
            hg = torch.cat([global_add_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)

        hg = F.relu(self.fc1(hg))
        out = self.fc2(hg)

        return out


# Tian Xie and Jeffrey C. Grossman
# Crystal Graph Convolutional Neural Networks for an Accurate and Interpretable Prediction of Material Properties
# Physical Review Letters, 120, 145301
class CGCNN(GNN):
    """
    Crystal graph convolutional neural network (CGCNN) form the `"Crystal Graph Convolutional Neural Networks for an
    Accurate and Interpretable Prediction of Material Properties"
    <https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.120.145301>`_ paper.
    """

    def __init__(self,
                 n_node_feats: int,
                 n_edge_feats: int,
                 dim_out: int,
                 n_graphs: int = 1,
                 readout: str = env.READOUT_MEAN):
        super(CGCNN, self).__init__(dim_out, env.ALG_CGCNN)
        self.n_graphs = n_graphs
        self.readout = readout
        self.fc1 = nn.Linear(n_node_feats, 256)
        self.gc1 = CGConv(256, n_edge_feats)
        self.gn1 = LayerNorm(256)
        self.gc2 = CGConv(256, n_edge_feats)
        self.gn2 = LayerNorm(256)
        self.gc3 = CGConv(256, n_edge_feats)
        self.gn3 = LayerNorm(256)
        self.fc2 = nn.Linear(self.n_graphs * 256, 32)
        self.fc3 = nn.Linear(32, dim_out)

    def __emb_nodes(self,
                    g: torch_geometric.data.Batch):
        """
        Generate node embeddings using the graph convolutional layers.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Node embeddings.
        """
        h = F.relu(self.fc1(g.x))
        h = F.relu(self.gn1(self.gc1(h, g.edge_index, g.edge_attr)))
        h = F.relu(self.gn2(self.gc2(h, g.edge_index, g.edge_attr)))
        h = F.relu(self.gn3(self.gc3(h, g.edge_index, g.edge_attr)))

        return h

    def forward(self,
                g: torch_geometric.data.Batch):
        """
        Predict target values for the given Batch object.

        :param g: (torch_geometric.data.Batch) An input Batch object of the torch_geometric.data.Data objects.
        :return: Target values.
        """

        if self.readout == env.READOUT_MEAN:
            hg = torch.cat([global_mean_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)
        elif self.readout == env.READOUT_SUM:
            hg = torch.cat([global_add_pool(self.__emb_nodes(b), b.batch) for b in g], dim=1)

        hg = F.relu(self.fc2(hg))
        out = self.fc3(hg)

        return out



"""
Base Utilities
--------------
The ``maica_old.chem.base`` module includes basic information and utilities to process chemical data.
This module was designed to provide elemental attributes or embeddings
in converting unstructured chemical data into the formal data objects such as ``numpy.ndarray`` and ``torch.Tensor``.
"""


import numpy
import json
from mendeleev.fetch import fetch_table
from sklearn import preprocessing


# A dictionary of atomic number for each atomic symbol.
atom_nums = {'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
             'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20,
             'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30,
             'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40,
             'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50,
             'Sb': 51, 'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60,
             'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70,
             'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80,
             'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90,
             'Pa': 91, 'U': 92, 'Np': 93, 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100}

# A dictionary of atomic symbol for each atomic number.
atom_syms = {v: k for k, v in atom_nums.items()}

# Basic elemental features from the Python Mendeleev package.
elem_feat_names = ['atomic_number', 'period', 'en_pauling', 'covalent_radius_bragg',
                   'electron_affinity', 'atomic_volume', 'atomic_weight', 'fusion_heat']

# Categories of hybridization of the atoms.
cat_hbd = ['SP', 'SP2', 'SP3', 'SP3D', 'SP3D2']

# Categories of formal charge of the atoms.
cat_fc = ['-4', '-3', '-2', '-1', '0', '1', '2', '3', '4']

# Bond types.
cat_bond_types = ['UNSPECIFIED', 'SINGLE', 'DOUBLE', 'TRIPLE', 'QUADRUPLE', 'QUINTUPLE', 'HEXTUPLE', 'ONEANDAHALF',
                  'TWOANDAHALF', 'THREEANDAHALF', 'FOURANDAHALF', 'FIVEANDAHALF', 'AROMATIC', 'IONIC', 'HYDROGEN',
                  'THREECENTER', 'DATIVEONE', 'DATIVE', 'DATIVEL', 'DATIVER', 'OTHER', 'ZERO']

# First ionization energies for each element.
first_ion_energies = [1312, 2372.3, 520.2, 899.5, 800.6, 1086.5, 1402.3, 1313.9, 1681, 2080.7,
                      495.8, 737.7, 577.5, 786.5, 1011.8, 999.6, 1251.2, 1520.6, 418.8, 589.8,
                      633.1, 658.8, 650.9, 652.9, 717.3, 762.5, 760.4, 737.1, 745.5, 906.4,
                      578.8, 762, 947, 941, 1139.9, 1350.8, 403, 549.5, 600, 640.1,
                      652.1, 684.3, 702, 710.2, 719.7, 804.4, 731, 867.8, 558.3, 708.6,
                      834, 869.3, 1008.4, 1170.4, 375.7, 502.9, 538.1, 534.4, 527, 533.1,
                      540, 544.5, 547.1, 593.4, 565.8, 573, 581, 589.3, 596.7, 603.4,
                      523.5, 658.5, 761, 770, 760, 840, 880, 870, 890.1, 1007.1,
                      589.4, 715.6, 703, 812.1, 899.003, 1037, 380, 509.3, 499, 587,
                      568, 597.6, 604.5, 584.7, 578, 581, 601, 608, 619, 627,
                      635, 642, 470, 580, 665, 757, 740, 730, 800, 960,
                      1020, 1155, 707.2, 832.2, 538.3, 663.9, 736.9, 860.1, 463.1, 563.3]


def load_mendeleev_feats():
    """
    Load elemental features from the `Python Mendeleev package <https://mendeleev.readthedocs.io/en/stable/>`_.
    Total eight elemental features containing the attributes in ``elem_feat_names`` and the first ionization energy
    are returned as ``numpy.ndarray`` for each element.
    The elemental features are normalized by z-score.

    :return: (*numpy.ndarray*) Pre-defined elemental features for each element in row-wise.
    """

    # Elemental attributes from the Python Mendeleev package.
    elem_feats = numpy.nan_to_num(numpy.array(fetch_table('elements')[elem_feat_names]))

    # First ionization energies of the elements.
    ion_energies = numpy.array(first_ion_energies[:elem_feats.shape[0]]).reshape(-1, 1)

    # Return normalized elemental features.
    return preprocessing.scale(numpy.hstack([elem_feats, ion_energies]))


def load_elem_feats(path_elem_embs: str = None):
    """
    Load elemental features from user-defined elemental embeddings in ``path_elem_embs``.
    The users can use the customized elemental features in machine learning by calling this function.
    The user-defined elemental features should be provided as a :obj:`JSON file` with a format of :obj:`{element: array of features}`.
    If the argument ``path_elem_embs`` is ``None``, basic elemental features from
    the `Python Mendeleev package <https://mendeleev.readthedocs.io/en/stable/>`_ is returned.

    :param path_elem_embs: (*str, optional*) Path of the JSON file including custom elemental features (*default* = ``None``).
    :return: (*numpy.ndarray*) Pre-defined elemental features or user-defined elemental embeddings for each element in row-wise.
    """

    if path_elem_embs is None:
        # Return basic elemental features from the Mendeleev package.
        return load_mendeleev_feats()
    else:
        # Get user-defined elemental features from a given elemental embeddings.
        elem_feats = list()
        with open(path_elem_embs) as json_file:
            elem_embs = json.load(json_file)
            for elem in atom_nums.keys():
                elem_feats.append(numpy.array(elem_embs[elem]))

        return numpy.vstack(elem_feats)



"""
Base Module
-----------
This module contains an abstract class of the dataset objects.
All dataset objects in MAICA inherits the abstract class of this module.
"""


import numpy
import pandas
import warnings
import copy
from abc import ABC
from abc import abstractmethod
from maica.core.sys import *


# Ignore unnecessary warnings from third party packages.
warnings.filterwarnings(action='ignore', category=UserWarning)


class Dataset(ABC):
    """
    An abstract class of Dataset classes in MAICA.
    The dataset objects in MAICA must inherit this class.
    """

    @abstractmethod
    def __init__(self,
                 data: object,
                 idx_feat: object,
                 idx_target: int,
                 var_names: numpy.ndarray):
        """
        :param data: (*object*) A data object.
        :param idx_feat: (*object*) (*numpy.ndarray*) Indices of the numerical features in the dataset file.
        :param idx_target: (*int*) An index of the target value in the dataset file.
        :param var_names: (*numpy.ndarray*) (*numpy.ndarray*) Names of all variables in the dataset file.
        """

        # Metadata of the input features and the target.
        self.idx_feat = copy.deepcopy(idx_feat)
        self.idx_target = copy.deepcopy(idx_target)
        self.contain_target = False if self.idx_target is None else True
        self.var_names = copy.deepcopy(var_names)
        self.feat_names = list()
        self.feat_types = list()
        self.target_name = None if self.idx_target is None else self.var_names[idx_target]

        # Data object.
        self.data = data

        # Tooltips for data visualization.
        self.tooltips = list()

        # Data statistics for normalization.
        self.feat_means = None
        self.feat_stds = None

    @abstractmethod
    def _set_feat_names(self):
        pass

    @abstractmethod
    def _set_tooltips(self):
        pass


def read_data_file(path_data_file: str):
    """
    Read data file form a given path of ``path_data_file``.
    Only the :obj:`.xlsx` and :obj:`.csv` extensions are available as the data file.

    :param path_data_file: (*str*) A path of the data file.
    :return: (*pandas.DataFrame, numpy.ndarray*) Data frame containing metadata and pure data object.
    """

    # Get the file extension of the data file.
    ext = path_data_file.split('.')[-1]

    # Read dataset file according to the file extension.
    if ext == 'xlsx':
        data_file = pandas.read_excel(path_data_file)
        data_obj = numpy.array(data_file)
    elif ext == 'csv':
        data_file = pandas.read_csv(path_data_file)
        data_obj = numpy.array(data_file)
    else:
        raise AssertionError(ERR_DATA_FILE_EXT.format(ext))

    return data_file, data_obj



"""
Neural Networks
---------------
The ``maica.ml.nn`` module provides an implementation of the most essential feedforward neural network.
The algorithms in this module are used to predict target values from the feature vectors and the chemical formulas.
"""


import numpy
import torch
import torch.nn as nn
import torch.nn.functional as F
import maica.core.sys as sys
import maica.core.env as env
from maica.ml.base import PyTorchModel


class FCNN(PyTorchModel):
    """
    Fully-connected neural network with the three hidden layers and the one output layer.
    For each hidden layer, the batch normalization technique is applied to accelerate the model training.
    """

    def __init__(self,
                 dim_in: int,
                 dim_out: int):
        super(FCNN, self).__init__(env.ALG_FCNN)
        self.fc1 = nn.Linear(dim_in, 256)
        self.bn1 = nn.BatchNorm1d(256)
        self.fc2 = nn.Linear(256, 256)
        self.bn2 = nn.BatchNorm1d(256)
        self.fc3 = nn.Linear(256, 32)
        self.bn3 = nn.BatchNorm1d(32)
        self.fc4 = nn.Linear(32, dim_out)

    def forward(self,
                x: torch.tensor):
        """
        Predict target values for the given data ``x``.

        :param x: (*torch.tensor*) A tensor containing input data of the model.
        :return: (*torch.Tensor*) Predicted values.
        """

        h = F.relu(self.bn1(self.fc1(x)))
        h = F.relu(self.bn2(self.fc2(h)))
        h = F.relu(self.bn3(self.fc3(h)))
        out = self.fc4(h)

        return out

    def fit(self,
            data_loader: torch.utils.data.DataLoader,
            optimizer: torch.optim.Optimizer,
            criterion: object):
        """
        Fit the model parameters for the given dataset using the data loader, the optimizer, and the loss function.
        It iterates the parameter optimization once for the entire dataset.

        :param data_loader: (*torch.utils.data.DataLoader*) A data loader to sample the data from the training dataset.
        :param optimizer: (*torch.optim.Optimizer*) An optimizer to fit the model parameters.
        :param criterion: (*object*) A loss function to evaluate the prediction performance of the model.
        :return: (*float*) Training loss.
        """

        self.train()
        sum_losses = 0

        for data, targets in data_loader:
            if sys.run_gpu:
                data = data.cuda()
                targets = targets.cuda()

            preds = self(data)
            loss = criterion(preds, targets.reshape(-1, 1))

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            sum_losses += loss.item()

        return sum_losses / len(data_loader)

    def predict(self,
                data_loader: object):
        """
        Predict target values for the given dataset in the data loader.

        :param data_loader: (*torch.utils.data.DataLoader*) A data loader to sample the data from the dataset.
        :return: (*numpy.ndarray*) Predicted values.
        """

        self.eval()
        list_preds = list()

        with torch.no_grad():
            for data in data_loader:
                if sys.run_gpu:
                    list_preds.append(self(data[0].cuda()).cpu().numpy())
                else:
                    list_preds.append(self(data[0]).numpy())

        return numpy.vstack(list_preds).flatten()


class Autoencoder(PyTorchModel):
    """
    A neural network to generate latent embeddings of input data.
    It is trained to minimize the difference between the input and its output rather than to be trained based on the target values.
    The training problem of the autoencoder can be defined by:

    .. math::
        \theta^* = \argmin_{\theta} ||\mathbf{x} - \mathbf{x}^{'}||_2^2,

    where :math:`mathbf{x}` is the input data, and :math:`\mathbf{x}^{'}` is the predicted value of the autoencoder.
    """

    def __init__(self,
                 dim_in: int,
                 dim_latent: int):
        super(Autoencoder, self).__init__(env.ALG_ATE)
        self.enc_fc1 = nn.Linear(dim_in, 256)
        self.enc_fc2 = nn.Linear(256, dim_latent)
        self.dec_fc1 = nn.Linear(dim_latent, 256)
        self.dec_fc2 = nn.Linear(256, dim_in)

    def forward(self,
                x: torch.tensor):
        """
        Perform encoding and decoding for the given data ``x``.

        :param x: (*torch.tensor*) The input data of the model.
        :return: The decoded input.
        """

        z = self.enc(x)
        x_p = self.dec(z)

        return x_p

    def enc(self,
            x: torch.tensor):
        """
        Generate the latent embedding of the input data ``x``.
        This is called :obj:`encoding` in the autoencoders.

        :param x: (*torch.tensor*) The input data of the model.
        :return: Latent embedding of the input data ``x``.
        """

        h = F.leaky_relu(self.enc_fc1(x))
        z = F.leaky_relu(self.enc_fc2(h))

        return z

    def dec(self,
            z: torch.tensor):
        """
        Restore the input data from the latent embedding of the input data.
        This is called :obj:`decoding` in the autoencoders.

        :param z: (*torch.tensor*) The latent embedding.
        :return: Restored input data from the given latent embedding ``z``.
        """

        h = F.leaky_relu(self.dec_fc1(z))
        x_p = self.dec_fc2(h)

        return x_p

    def fit(self,
            data_loader: torch.utils.data.DataLoader,
            optimizer: torch.optim.Optimizer):
        """
        Fit the model parameters for the given dataset using the data loader and the optimizer.
        It iterates the parameter optimization once for the entire dataset.

        :param data_loader: (*torch.utils.data.DataLoader*) A data loader to sample the data from the training dataset.
        :param optimizer: (*torch.optim.Optimizer*) An optimizer to fit the model parameters.
        :return: Training loss.
        """

        self.train()
        sum_losses = 0

        for data in data_loader:
            data = data[0]

            if sys.run_gpu:
                data = data.cuda()

            preds = self(data)
            loss = torch.mean((data - preds)**2)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            sum_losses += loss.item()

        return sum_losses / len(data_loader)

    def predict(self,
                data_loader: object):
        """
        Predict target values for the given dataset in the data loader.

        :param data_loader: (*torch.utils.data.DataLoader*) A data loader to sample the data from the dataset.
        :return: A NumPy array containing the predicted values.
        """

        self.eval()
        list_preds = list()

        with torch.no_grad():
            for data in data_loader:
                if sys.run_gpu:
                    list_preds.append(self.enc(data.cuda()).cpu().numpy())
                else:
                    list_preds.append(self.enc(data).numpy())

        return numpy.vstack(list_preds)



"""
Transfer Learning
-----------------
This module includes several functions to perform transfer learning using neural networks.
Popular transfer learning methods called :obj:`retrain head` and :obj:`fine tuning` are provided as built-in functions.
"""

import numpy
import torch
import maica_old.ml.base
import maica_old.data.base
import maica_old.core.env as env
import maica_old.ml.util as mu
from maica_old.data.vector import VectorDataset
from maica_old.data.graph import GraphDataset


def tl_retrain_head(path_source_model,
                    model_target: maica_old.ml.base.Model,
                    dataset_target: maica_old.data.base.Dataset,
                    gd_name: str = env.GD_ADAM,
                    init_lr: float = 1e-3,
                    l2_reg: float = 1e-6,
                    loss_func: str = env.LOSS_MAE,
                    max_epoch: int = 300):
    """
    Perform transfer learning based on feature extractor of the source model.
    The feature extraction layers of the source model are frozen in the training on the target dataset.
    Only the prediction layer called head is trained during the training on the target dataset.
    This transfer learning method is called 'retrain head'.

    :param path_source_model: (str) The path of the model file of the source model.
    :param model_target: (ml.base.Model) Target model that will be used for transfer learning.
    :param dataset_target: (data.base.Dataset) Target dataset for transfer learning.
    :param gd_name: (str, optional) A name of the optimizer to train the target model (default = AdamOptimizer)
    :param init_lr: (float, optional) An initial learning rate of the gradient descent optimizer (default = 1e-3).
    :param l2_reg: (float, optional) A coefficient of the L2 regularization in model parameters (Default = 1e-6).
    :param loss_func: (str, optional) A name of loss function to evaluate the model performance (default = Mean Absolute Error).
    :param max_epoch: (int, optional) The maximum iteration of the model parameter optimization in the training.
    :return: Trained model.
    """

    # Load model parameters of the source model into the target model.
    model_target.load(path_source_model)

    # Randomly initialize the model parameters of the last layer.
    head = list(model_target.children())[-1]
    torch.nn.init.normal_(head.weight)
    if head.bias is not None:
        torch.nn.init.normal_(head.bias)

    # Generate data loader of the target dataset.
    batch_size = int(numpy.minimum(64, dataset_target.n_data()))
    if isinstance(dataset_target, VectorDataset):
        data_loader = mu.get_data_loader(dataset_target.x(), dataset_target.y(), batch_size=batch_size, shuffle=True)
    elif isinstance(dataset_target, GraphDataset):
        data_loader = mu.get_data_loader(dataset_target, batch_size=batch_size, shuffle=True)

    # Get optimizer and loss function for the training.
    optimizer = mu.get_optimizer(head.parameters(), gd_name=gd_name, init_lr=init_lr, l2_reg=l2_reg)
    loss_func = mu.get_loss_func(loss_func=loss_func)

    # Optimize model parameters of the neural network.
    for i in range(0, max_epoch):
        train_loss = model_target.fit(data_loader, optimizer, loss_func)
        print('Epoch [{}/{}]\tTrain loss: {:.4f}'.format(i + 1, max_epoch, train_loss))

    return model_target


def tl_fine_tuning(path_source_model,
                   model_target: maica_old.ml.base.Model,
                   dataset_target: maica_old.data.base.Dataset,
                   gd_name: str = env.GD_ADAM,
                   init_lr: float = 1e-6,
                   l2_reg: float = 1e-6,
                   loss_func: str = env.LOSS_MAE,
                   max_epoch: int = 300):
    """
    Perform transfer learning based on a pre-trained source model.
    The target model is initialized by the model parameters of the source model.
    After the initialization, the target model is trained on the target dataset with a small learning rate.
    This transfer learning method is called 'fine tuning'.

    :param path_source_model: (str) The path of the model file of the source model.
    :param model_target: (ml.base.Model) Target model that will be used for transfer learning.
    :param dataset_target: (data.base.Dataset) Target dataset for transfer learning.
    :param gd_name: (str, optional) A name of the optimizer to train the target model (default = AdamOptimizer)
    :param init_lr: (float, optional) An initial learning rate of the gradient descent optimizer (default = 1e-3).
    :param l2_reg: (float, optional) A coefficient of the L2 regularization in model parameters (Default = 1e-6).
    :param loss_func: (str, optional) A name of loss function to evaluate the model performance (default = Mean Absolute Error).
    :param max_epoch: (int, optional) The maximum iteration of the model parameter optimization in the training.
    :return: Trained model.
    """

    # Load model parameters of the source model into the target model.
    model_target.load(path_source_model)

    # Generate data loader of the target dataset.
    batch_size = int(numpy.minimum(64, dataset_target.n_data()))
    if isinstance(dataset_target, VectorDataset):
        data_loader = mu.get_data_loader(dataset_target.x(), dataset_target.y(), batch_size=batch_size, shuffle=True)
    elif isinstance(dataset_target, GraphDataset):
        data_loader = mu.get_data_loader(dataset_target, batch_size=batch_size, shuffle=True)

    # Get optimizer and loss function for the training.
    optimizer = mu.get_optimizer(model_target.parameters(), gd_name=gd_name, init_lr=init_lr, l2_reg=l2_reg)
    loss_func = mu.get_loss_func(loss_func=loss_func)

    # Optimize model parameters of the neural network.
    for i in range(0, max_epoch):
        train_loss = model_target.fit(data_loader, optimizer, loss_func)
        print('Epoch [{}/{}]\tTrain loss: {:.4f}'.format(i + 1, max_epoch, train_loss))

    return model_target



"""
Environment Variables
---------------------
It is a core module of MAICA to configure framework environments.
This module contains environment variables to define prediction task, machine learning algorithm,
and other algorithmic descriptors.
"""


# Data Formats
DATA_FORMAT_VEC = 'vec'         # Numerical Vector
DATA_FORMAT_IMG = 'img'         # Image Data
DATA_FORMAT_GRAPH = 'graph'     # Mathematical Graph


# Data Types
DATA_TYPE_VEC = 'vec'           # Numerical Vector
DATA_TYPE_FORM = 'form'         # Chemical Formula
DATA_TYPE_SMILES = 'smiles'     # Simplified Molecular-Input Line-Entry System (SMILES)
DATA_TYPE_CIF = 'cif'           # Crystallographic Information File


# Feature Types
FEAT_TYPE_NUM = 'num'           # Numerical Value
FEAT_TYPE_FORM = 'form'         # Chemical Formula
FEAT_TYPE_STRUCT = 'struct'     # Mathematical Graph


# Imputation Methods
IMPUTE_MEAN = 'mean'            # Fill empty value by mean.
IMPUTE_ZERO = 'zero'            # Fill empty value by zero.
IMPUTE_KNN = 'knn'              # Fill empty value by k-nearest neighbor data.


# Abbreviations of the Machine Learning Algorithms
ALG_SKLEARN = 'sklearn'         # Abstract Algorithm of Scikit-learn
ALG_PYTORCH = 'pytorch'         # Abstract Algorithm of PyTorch
ALG_LR = 'lr'                   # Linear Regression
ALG_LASSO = 'lasso'             # Lasso
ALG_DCTR = 'dctr'               # Decision Tree Regression
ALG_SYMR = 'symr'               # Symbolic Regression
ALG_KRR = 'krr'                 # Kernel Ridge Regression
ALG_KNNR = 'knnr'               # K-Nearest Neighbor Regression
ALG_SVR = 'svr'                 # Support Vector Regression
ALG_GPR = 'gpr'                 # Gaussian Process Regression
ALG_GBTR = 'gbtr'               # Gradient Boosting Tree Regression
ALG_FCNN = 'fcnn'               # Fully-Connected Neural Network
ALG_ATE = 'autoencoder'         # Autoencoder
ALG_DOPNET = 'dopnet'           # DopNet
ALG_GCN = 'gcn'                 # Graph Convolutional Network
ALG_GAT = 'gat'                 # Graph Attention Network
ALG_GIN = 'gin'                 # Graph Isomorphism Network
ALG_CGCNN = 'cgcnn'             # Crystal Graph Convolutional Neural Network


# Gradient-Based Optimizers
GD_SGD = 'sgd'                  # Stochastic Gradient Descent Method
GD_ADADELTA = 'adadelta'        # AdaDelta Optimizer
GD_RMSPROP = 'rmsprop'          # Root Mean Square Propagation
GD_ADAM = 'adam'                # Adam Optimizer


# Loss Function
LOSS_MAE = 'mae'                # Mean Absolute Error
LOSS_MSE = 'mse'                # Mean Squared Error
Loss_SMAE = 'smae'              # Smooth Mean Absolute Error


# Readout Methods for Graph Neural Networks
READOUT_MEAN = 'mean'           # Mean-Based Readout
READOUT_SUM = 'add'             # Sum-Based Readout


# Manifold Learning Methods
EMB_TSNE = 'tsne'               # t-Distributed Stochastic Neighbor Embedding
EMB_SPECT = 'spect'             # Spectral Embedding


# Available Algorithms for Each Pipeline
ALGS_SKLEARN = [ALG_SKLEARN, ALG_LR, ALG_LASSO, ALG_DCTR, ALG_SYMR, ALG_KRR, ALG_KNNR, ALG_SVR, ALG_GPR, ALG_GBTR]
ALGS_PYTORCH = [ALG_PYTORCH, ALG_FCNN, ALG_DOPNET, ALG_GCN, ALG_GAT, ALG_GIN, ALG_CGCNN]


# Optimization Types
PROB_MIN = 'min'                # Minimization Problem
PROB_MAX = 'max'                # Maximization Problem


# Meta-heuristic Algorithms
OPT_EP = 'ep'                   # Evolutionary Programming
OPT_ES = 'es'                   # Evolution Strategies
OPT_MA = 'ma'                   # Memetic Algorithm
OPT_GA = 'ga'                   # Genetic Algorithm
OPT_DE = 'de'                   # Differential Evolution
OPT_FPA = 'fpa'                 # Flower Pollination Algorithm
OPT_CRO = 'cro'                 # Coral Reefs Optimization
OPT_PSO = 'pso'                 # Particle Swarm Optimization
OPT_BFO = 'bfo'                 # Bacterial Foraging Optimization
OPT_BEES = 'bees'               # Bees Algorithm
OPT_CSO = 'cso'                 # Cat Swarm Optimization
OPT_ACO = 'aco'                 # Ant Colony Optimization
OPT_ABC = 'abc'                 # Artificial Bee Colony
OPT_CSA = 'csa'                 # Cuckoo Search Algorithm
OPT_FFA = 'ffa'                 # Firefly Algorithm
OPT_FA = 'fa'                   # Fireworks Algorithm
OPT_BA = 'ba'                   # Bat Algorithm
OPT_FOA = 'foa'                 # Fruit-fly Optimization Algorithm
OPT_SSO = 'sso'                 # Social Spider Optimization
OPT_GWO = 'gwo'                 # Grey Wolf Optimizer
OPT_SSA = 'ssa'                 # Social Spider Algorithm
OPT_ALO = 'alo'                 # Ant Lion Optimizer
OPT_MFO = 'mfo'                 # Moth Flame Optimization
OPT_EHO = 'eho'                 # Elephant Herding Optimization
OPT_JA = 'ja'                   # Jaya Algorithm
OPT_WOA = 'woa'                 # Whale Optimization Algorithm
OPT_DO = 'do'                   # Dragonfly Optimization
OPT_BSA = 'bsa'                 # Bird Swarm Algorithm
OPT_SHO = 'sho'                 # Spotted Hyena Optimizer
OPT_SALP = 'salp'               # Salp Swarm Optimization
OPT_SRSR = 'srsr'               # Swarm Robotics Search And Rescue
OPT_GOA = 'goa'                 # Grasshopper Optimization Algorithm
OPT_MSA = 'msa'                 # Moth Search Algorithm
OPT_SLO = 'slo'                 # Sea Lion Optimization
OPT_NMRA = 'nmra'               # Nake Mole-rat Algorithm
OPT_BES = 'bes'                 # Bald Eagle Search
OPT_PFA = 'pfa'                 # Pathfinder Algorithm
OPT_SFO = 'sfo'                 # Sailfish Optimizer
OPT_HHO = 'hho'                 # Harris Hawks Optimization
OPT_MRFO = 'mrfo'               # Manta Ray Foraging Optimization
OPT_SPA = 'spa'                 # Sparrow Search Algorithm
OPT_HGS = 'hgs'                 # Hunger Games Search
OPT_AO = 'ao'                   # Aquila Optimizer
OPT_SA = 'sa'                   # Simulated Annealing
OPT_WDO = 'wdo'                 # Wind Driven Optimization
OPT_MVO = 'mvo'                 # Multi-Verse Optimizer
OPT_TWO = 'two'                 # Tug of War Optimization
OPT_EFO = 'efo'                 # Electromagnetic Field Optimization
OPT_NRO = 'nro'                 # Nuclear Reaction Optimization
OPT_HGSO = 'hgso'               # Henry Gas Solubility Optimization
OPT_ASO = 'aso'                 # Atom Search Optimization
OPT_EO = 'eo'                   # Equilibrium Optimizer
OPT_ARCH = 'arch'               # Archimedes Optimization Algorithm
OPT_CA = 'ca'                   # Culture Algorithm
OPT_ICA = 'ica'                 # Imperialist Competitive Algorithm
OPT_TLO = 'tlo'                 # Teaching Learning Optimization
OPT_BSO = 'bso'                 # Brain Storm Optimization
OPT_QSA = 'qsa'                 # Queuing Search Algorithm
OPT_SARO = 'saro'               # Search And Rescue Optimization
OPT_LCBO = 'lcbo'               # Life Choice-Based Optimization
OPT_SSDO = 'ssdo'               # Social Ski-Driver Optimization
OPT_GSKA = 'gska'               # Gaining Sharing Knowledge-based Algorithm
OPT_CHIO = 'chio'               # Coronavirus Herd Immunity Optimization
OPT_FBIO = 'fbio'               # Forensic-Based Investigation Optimization
OPT_BRO = 'bro'                 # Battle Royale Optimization
OPT_IWO = 'iwo'                 # Invasive Weed Optimization
OPT_BBO = 'bbo'                 # Biogeography-Based Optimization
OPT_VCS = 'vcs'                 # Virus Colony Search
OPT_SBO = 'sbo'                 # Satin Bowerbird Optimizer
OPT_EOA = 'eoa'                 # Earthworm Optimisation Algorithm
OPT_WHO = 'who'                 # Wildebeest Herd Optimization
OPT_SMA = 'sma'                 # Slime Mould Algorithm
OPT_GCO = 'gco'                 # Germinal Center Optimization
OPT_WCA = 'wca'                 # Water Cycle Algorithm
OPT_AEO = 'aeo'                 # Artificial Ecosystem-based Optimization
OPT_HC = 'hc'                   # Hill Climbing
OPT_SCA = 'sca'                 # Sine Cosine Algorithm
OPT_AOA = 'aoa'                 # Arithmetic Optimization Algorithm
OPT_HS = 'hs'                   # Harmony Search
OPT_CEM = 'cem'                 # Cross-Entropy Method
OPT_PIO = 'pio'                 # Pigeon-Inspired Optimization
OPT_AAA = 'aaa'                 # Artificial Algae Algorithm
OPT_RHO = 'rho'                 # Rhino Herd Optimization
OPT_EPO = 'epo'                 # Emperor Penguin Optimizer
OPT_BOA = 'boa'                 # Butterfly Optimization Algorithm
OPT_BMO = 'bmo'                 # Blue Monkey Optimization
OPT_SOA = 'soa'                 # Sandpiper Optimization Algorithm
OPT_BWO = 'bwo'                 # Black Widow Optimization



import numpy
import torch
import torch.nn.functional as F
from torch.utils.data import TensorDataset


def train_emb_net(emb_net, data_loader, optimizer, max_epoch):
    if isinstance(data_loader.dataset, TensorDataset):
        for i in range(0, max_epoch):
            train_loss = __train_vec(emb_net, data_loader, optimizer)
            print('Epoch [{}/{}]\tTrain loss: {:.4f}'.format(i + 1, max_epoch, train_loss))
    elif isinstance(data_loader.dataset, list):
        for i in range(0, max_epoch):
            print('graph')
            train_loss = __train_graph(emb_net, data_loader, optimizer)
            print('Epoch [{}/{}]\tTrain loss: {:.4f}'.format(i + 1, max_epoch, train_loss))
    return None


def __train_vec(emb_net, data_loader, optimizer):
    emb_net.train()
    train_loss = 0

    for data, targets in data_loader:
        data = data.cuda()
        targets = targets.cuda()

        rand_idx = torch.randperm(data.shape[0])
        inputs_sample = data[rand_idx, :]
        targets_sample = targets[rand_idx, :]

        emb_data = F.normalize(emb_net(data), p=2, dim=1)
        emb_sample = F.normalize(emb_net(inputs_sample), p=2, dim=1)

        dist_emb = torch.sum((emb_data - emb_sample)**2, dim=1)
        dist_target = torch.sum((targets - targets_sample)**2, dim=1)
        loss = torch.mean((dist_emb - dist_target)**2)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        train_loss += loss.item()

    return train_loss / len(data_loader)


def __train_graph():
    return None


def test(emb_net, data_loader):
    emb_net.eval()
    list_embs = list()

    with torch.no_grad():
        for data, _ in data_loader:
            data = data.cuda()

            embs = F.normalize(emb_net(data), p=2, dim=1)
            list_embs.append(embs.cpu().numpy())

    return numpy.vstack(list_embs)



"""
Optimization
------------
The ``maica_old.util.optimization`` module supports the function optimization.
It includes various meta-heuristic algorithms and a wrapper function to execute them.
This module can be used to optimize the input data of user-defined functions or machine learning models.
"""


import numpy
import torch
from mealpy.evolutionary_based.EP import BaseEP
from mealpy.evolutionary_based.ES import BaseES
from mealpy.evolutionary_based.MA import BaseMA
from mealpy.evolutionary_based.GA import BaseGA
from mealpy.evolutionary_based.DE import BaseDE
from mealpy.evolutionary_based.FPA import BaseFPA
from mealpy.evolutionary_based.CRO import BaseCRO
from mealpy.swarm_based.PSO import BasePSO
from mealpy.swarm_based.BFO import BaseBFO
from mealpy.swarm_based.BeesA import BaseBeesA
from mealpy.swarm_based.CSO import BaseCSO
from mealpy.swarm_based.ACOR import BaseACOR
from mealpy.swarm_based.ABC import BaseABC
from mealpy.swarm_based.CSA import BaseCSA
from mealpy.swarm_based.FireflyA import BaseFireflyA
from mealpy.swarm_based.FA import BaseFA
from mealpy.swarm_based.BA import BaseBA
from mealpy.swarm_based.FOA import BaseFOA
from mealpy.swarm_based.SSO import BaseSSO
from mealpy.swarm_based.GWO import BaseGWO
from mealpy.swarm_based.SSA import BaseSSA
from mealpy.swarm_based.ALO import BaseALO
from mealpy.swarm_based.MFO import BaseMFO
from mealpy.swarm_based.EHO import BaseEHO
from mealpy.swarm_based.JA import BaseJA
from mealpy.swarm_based.WOA import BaseWOA
from mealpy.swarm_based.DO import BaseDO
from mealpy.swarm_based.BSA import BaseBSA
from mealpy.swarm_based.SHO import BaseSHO
from mealpy.swarm_based.SalpSO import BaseSalpSO
from mealpy.swarm_based.SRSR import BaseSRSR
from mealpy.swarm_based.GOA import BaseGOA
from mealpy.swarm_based.MSA import BaseMSA
from mealpy.swarm_based.NMRA import ImprovedNMR
from mealpy.swarm_based.BES import BaseBES
from mealpy.swarm_based.PFA import ImprovedPFA
from mealpy.swarm_based.SFO import ImprovedSFO
from mealpy.swarm_based.HHO import BaseHHO
from mealpy.swarm_based.MRFO import BaseMRFO
from mealpy.swarm_based.SpaSA import BaseSpaSA
from mealpy.swarm_based.HGS import OriginalHGS
from mealpy.physics_based.SA import BaseSA
from mealpy.physics_based.WDO import BaseWDO
from mealpy.physics_based.MVO import BaseMVO
from mealpy.physics_based.TWO import ImprovedTWO
from mealpy.physics_based.EFO import BaseEFO
from mealpy.physics_based.NRO import BaseNRO
from mealpy.physics_based.HGSO import BaseHGSO
from mealpy.physics_based.ASO import BaseASO
from mealpy.physics_based.EO import BaseEO
from mealpy.human_based.CA import OriginalCA
from mealpy.human_based.ICA import BaseICA
from mealpy.human_based.TLO import BaseTLO
from mealpy.human_based.BSO import BaseBSO
from mealpy.human_based.QSA import BaseQSA
from mealpy.human_based.SARO import BaseSARO
from mealpy.human_based.LCBO import BaseLCBO
from mealpy.human_based.SSDO import BaseSSDO
from mealpy.human_based.GSKA import BaseGSKA
from mealpy.human_based.CHIO import BaseCHIO
from mealpy.human_based.FBIO import BaseFBIO
from mealpy.human_based.BRO import BaseBRO
from mealpy.bio_based.IWO import BaseIWO
from mealpy.bio_based.BBO import BaseBBO
from mealpy.bio_based.VCS import BaseVCS
from mealpy.bio_based.SBO import BaseSBO
from mealpy.bio_based.EOA import BaseEOA
from mealpy.bio_based.WHO import BaseWHO
from mealpy.bio_based.SMA import BaseSMA
from mealpy.system_based.GCO import BaseGCO
from mealpy.system_based.WCA import BaseWCA
from mealpy.system_based.AEO import BaseAEO
from mealpy.math_based.HC import BaseHC
from mealpy.math_based.SCA import BaseSCA
from mealpy.music_based.HS import BaseHS
from mealpy.probabilistic_based.CEM import BaseCEM
from mealpy.dummy.PIO import BasePIO
from mealpy.dummy.AAA import BaseAAA
from mealpy.dummy.RHO import BaseRHO
from mealpy.dummy.EPO import BaseEPO
from mealpy.dummy.BOA import BaseBOA
from mealpy.dummy.BMO import BaseBMO
from mealpy.dummy.SOA import BaseSOA
from mealpy.dummy.BWO import BaseBWO
from maica_old.ml.base import Model
from maica_old.core.env import *


def optimize(obj_func: object,
             opt_alg: str,
             opt_type: str,
             return_hist: bool = None,
             **kwargs):
    """
    Optimize the given user-defined function ``obj_func`` using the meta-heuristic algorithm ``opt_alg`` in ARTCAI.
    The objective function can be all input-output mappings regardless of their differentiability.

    :param obj_func: (*object*) The objective function to optimize.
    :param opt_alg: (*str*) A name of the meta-heuristic algorithm to optimize the objective function.
    :param opt_type: (*str*) An argument for determining the objective function to be minimized or maximized.
    :param return_hist: (*bool*) Return the optimized inputs over the iterations (*default* = ``None``).
    :param kwargs: (*optional*) Hyperparameters of the meta-heuristic algorithm.
    :return: Optimal input of the objective function and its score in the meta-heuristic algorithm.
    """

    if opt_type == PROB_MIN:
        # Objective function for minimization problem.
        __obj_func = obj_func
    elif opt_type == PROB_MAX:
        # Negative objective function for maximization problem.
        def __obj_func(x: numpy.ndarray):
            return -obj_func(x)
    else:
        raise AssertionError('Unknown optimization type is given: {}.'.format(opt_type))

    if opt_alg == OPT_EP:
        optimizer = BaseEP(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_ES:
        optimizer = BaseES(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_MA:
        optimizer = BaseMA(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_GA:
        optimizer = BaseGA(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_DE:
        optimizer = BaseDE(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_FPA:
        optimizer = BaseFPA(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_CRO:
        optimizer = BaseCRO(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_PSO:
        optimizer = BasePSO(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_BFO:
        optimizer = BaseBFO(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_BEES:
        optimizer = BaseBeesA(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_CSO:
        optimizer = BaseCSO(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_ACO:
        optimizer = BaseACOR(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_ABC:
        optimizer = BaseABC(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_CSA:
        optimizer = BaseCSA(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_FFA:
        optimizer = BaseFireflyA(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_FA:
        optimizer = BaseFA(obj_func=obj_func, **kwargs)
    elif opt_alg == OPT_BA:
        optimizer = BaseBA(obj_func=obj_func, **kwargs)

    best_sol, best_score, _ = optimizer.train()

    if return_hist:
        return best_sol, best_score, optimizer.hist
    else:
        return best_sol, best_score


def run_ml_model(model: Model,
                 input_data: numpy.ndarray):
    """
    Run the given machine learning model of ``model`` to generate model outputs.
    It is a wrapper function to use the machine learning models as an objective function of meta-heuristics.

    :param model: (*maica_old.ml.base.Model*) A machine learning model that makes up the objective function.
    :param input_data: (*numpy.ndarray*) Input data of the given machine learning model.
    :return: Output data of the machine learning model.
    """

    if model.alg_type == ALG_SKLEARN:
        return model.predict(input_data.reshape(1, -1))
    elif model.alg_type == ALG_PYTORCH:
        return model(torch.tensor(input_data, dtype=torch.float).view(1, -1))
    else:
        raise AssertionError('Unknown algorithm is given: {}.'.format(model.alg_type))



"""
Base Utilities
--------------
The ``maica_old.chem.base`` module includes basic information and utilities to process chemical data.
This module was designed to provide elemental attributes or embeddings
in converting unstructured chemical data into the formal data formats, such as numpy.ndarray and torch.Tensor.
"""


import numpy
import json
from mendeleev.fetch import fetch_table
from mendeleev import element
from sklearn import preprocessing


# A dictionary of atomic number for each atomic symbol.
atom_nums = {'H': 1, 'He': 2, 'Li': 3, 'Be': 4, 'B': 5, 'C': 6, 'N': 7, 'O': 8, 'F': 9, 'Ne': 10,
             'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14, 'P': 15, 'S': 16, 'Cl': 17, 'Ar': 18, 'K': 19, 'Ca': 20,
             'Sc': 21, 'Ti': 22, 'V': 23, 'Cr': 24, 'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29, 'Zn': 30,
             'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34, 'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39, 'Zr': 40,
             'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44, 'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49, 'Sn': 50,
             'Sb': 51, 'Te': 52, 'I': 53, 'Xe': 54, 'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59, 'Nd': 60,
             'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64, 'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69, 'Yb': 70,
             'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74, 'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79, 'Hg': 80,
             'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84, 'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89, 'Th': 90,
             'Pa': 91, 'U': 92, 'Np': 93, 'Pu': 94, 'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99, 'Fm': 100}

# A dictionary of atomic symbol for each atomic number.
atom_syms = {v: k for k, v in atom_nums.items()}

# Basic elemental features from the Python Mendeleev package.
elem_feat_names = ['atomic_number', 'period', 'en_pauling', 'covalent_radius_bragg',
                   'electron_affinity', 'atomic_volume', 'atomic_weight', 'fusion_heat']

# Categories of hybridization of the atoms.
cat_hbd = ['SP', 'SP2', 'SP3', 'SP3D', 'SP3D2']

# Categories of formal charge of the atoms.
cat_fc = ['-4', '-3', '-2', '-1', '0', '1', '2', '3', '4']

# Bond types.
cat_bond_types = ['UNSPECIFIED', 'SINGLE', 'DOUBLE', 'TRIPLE', 'QUADRUPLE', 'QUINTUPLE', 'HEXTUPLE', 'ONEANDAHALF',
                  'TWOANDAHALF', 'THREEANDAHALF', 'FOURANDAHALF', 'FIVEANDAHALF', 'AROMATIC', 'IONIC', 'HYDROGEN',
                  'THREECENTER', 'DATIVEONE', 'DATIVE', 'DATIVEL', 'DATIVER', 'OTHER', 'ZERO']


def load_mendeleev_feats():
    """
    Load elemental features from the `Python Mendeleev package <https://mendeleev.readthedocs.io/en/stable/>`_.
    Total eight elemental features containing the attributes in ``elem_feat_names`` and the first ionization energy are loaded.

    :return: An array of the basic elemental features in row-wise.
    """

    elem_feats = numpy.nan_to_num(numpy.array(fetch_table('elements')[elem_feat_names]))
    ion_engs = numpy.zeros((elem_feats.shape[0], 1))

    # Get first ionization energies for each element.
    for i in range(0, ion_engs.shape[0]):
        ion_eng = element(i + 1).ionenergies
        if 1 in ion_eng:
            ion_engs[i, 0] = element(i + 1).ionenergies[1]
        else:
            ion_engs[i, 0] = 0

    # Return normalized elemental features.
    return preprocessing.scale(numpy.hstack([elem_feats, ion_engs]))


def load_elem_feats(path_elem_embs: str = None):
    """
    Load elemental features from the elemental embeddings in the given ``path_elem_embs``.
    Users can use the customized elemental features in machine learning by exploiting this function.
    The user-defined elemental features should be provided as a JSON file with a format of {element: array of features}.
    If the argument ``path_elem_embs`` is ``None``, basic elemental features from
    the `Python Mendeleev package <https://mendeleev.readthedocs.io/en/stable/>`_ is returned.

    :param path_elem_embs: (*str, optional*) Path of the JSON file including custom elemental features (*default* = ``None``).
    :return: An array of the elemental features in row-wise.
    """

    if path_elem_embs is None:
        # Return basic elemental features from the Mendeleev package.
        return load_mendeleev_feats()
    else:
        # Get user-defined elemental features from the given elemental embeddings.
        elem_feats = list()
        with open(path_elem_embs) as json_file:
            elem_embs = json.load(json_file)
            for elem in atom_nums.keys():
                elem_feats.append(numpy.array(elem_embs[elem]))

        return numpy.vstack(elem_feats)



"""
Chemical Formula
----------------
A collection of functions to convert the string of the chemical formulas into
the machine-readable numerical representations.
"""


import numpy
import ast
from maica_old.chem.base import atom_nums


def form_to_vec(form: str,
                elem_feats: numpy.ndarray):
    """
    Convert a given chemical formula in ``form`` into a feature vector.
    For a set of computed atomic features :math:`S = \{\mathbf{h} = f(e) | e \in c \}`
    where :math:`c` is the given chemical formula, the feature vector is calculated as a concatenated vector based on
    weighted sum with a weight :math:`w_\mathbf{h}`, standard deviation :math:`\sigma`, and max operation as:

    .. math::
        \mathbf{x} = \sum_{\mathbf{h} \in S} w_\mathbf{h} h \oplus \sigma(S) \oplus \max(S).

    Note that the standard deviation :math:`\sigma` and the max operations are applied feature-wise (not element-wise).
    This formula-to-vector conversion method is common in chemical machine learning
    [`1 <https://pubs.acs.org/doi/abs/10.1021/acs.jpclett.8b00124>`_,
    `2 <https://journals.aps.org/prb/abstract/10.1103/PhysRevB.93.115104>`_,
    `3 <https://www.nature.com/articles/s41524-021-00564-y>`_].

    :param form: (*str*) Chemical formula (e.g., ZnIn2S4).
    :param elem_feats: (*numpy.ndarray*) The NumPy array of elemental features.
    :return: A feature vector of the chemical formula.
    """

    wt_sum_feats = numpy.zeros(elem_feats.shape[1])
    list_atom_feats = list()

    # Convert the refined chemical formula to a dictionary of the chemical formula.
    elems = ast.literal_eval(form)
    sum_elem_nums = numpy.sum([float(elems[e]) for e in elems])

    # Get atomic features for each element.
    for e in elems:
        atom_feats = elem_feats[atom_nums[e] - 1, :]
        list_atom_feats.append(atom_feats)
        wt_sum_feats += (float(elems[e]) / sum_elem_nums) * atom_feats

    # Generate a feature vector of the formula based on the weighted sum, std., and max. of the atomic features.
    form_vec = numpy.hstack([wt_sum_feats, numpy.std(list_atom_feats, axis=0), numpy.max(list_atom_feats, axis=0)])

    return form_vec


def form_to_sparse_vec(form: str,
                       elem_feats: numpy.ndarray,
                       max_elems: int):
    """
    Convert a given chemical formula in ``form`` into a feature vector.
    For a set of computed atomic features :math:`S = \{\mathbf{h} = f(e) | e \in c \}`
    where :math:`c` is the given chemical formula, the feature vector is calculated as a concatenated vector of the atomic features.
    If the number of elements is less than ``max_elems``, a zero vector is concatenated to fill the empty values.
    Thus, it always returns the feature vector with the same dimension based on ``max_elems``
    regardless of the number of elements in the chemical formula.

    :param form: (*str*) Chemical formula (e.g., ZnIn2S4).
    :param elem_feats: (*numpy.ndarray*) The NumPy array of elemental features.
    :param max_elems: (*int*) The maximum number of elements in the chemical formula.
    :return: A feature vector of the chemical formula.
    """

    # The number of features = the number of elemental features + ratio of the atom in the formula.
    form_feats = list()
    elems = ast.literal_eval(form)

    # Get atomic features for each element.
    for e in elems:
        form_feats.append(float(elems[e]))
        form_feats.append(elem_feats[atom_nums[e] - 1, :])

    # Fill zero vectors to the empty features.
    form_feats.append(numpy.zeros((max_elems - len(elems)) * (elem_feats.shape[1] + 1)))

    # Generate a sparse feature vector of the given chemical formula.
    form_feats = numpy.hstack(form_feats)

    return form_feats


def parse_form(form: str):
    """
    Parse the chemical formula to an element-wise dictionary.
    For example, ZnIn2S4 is converted into a dictionary of {Zn: 1.0, In: 2.0, S: 4.0}.

    :param form: (*str*) Chemical formula (e.g., ZnIn2S4).
    :return: An element-wise dictionary of the chemical formula.
    """

    form_dict = dict()
    elem = form[0]
    num = ''

    # Convert the chemical formula to a dictionary of the chemical formula.
    for i in range(1, len(form)):
        if form[i].islower():
            elem += form[i]
        elif form[i].isupper():
            if num == '':
                form_dict[elem] = 1.0
            else:
                form_dict[elem] = float(num)

            elem = form[i]
            num = ''
        elif form[i].isnumeric() or form[i] == '.':
            num += form[i]

        if i == len(form) - 1:
            if num == '':
                form_dict[elem] = 1.0
            else:
                form_dict[elem] = float(num)

    # Return a dictionary of {element: ratio} pairs.
    return form_dict



"""
Base Classes
------------
The ``maica.ml.base`` module includes basic classes of the machine learning algorithms.
It provides a wrapper class of the Scikit-learn models and an abstract class of the PyTorch models.
"""


import numpy
import joblib
from abc import ABC
from abc import abstractmethod
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Lasso
from sklearn.tree import DecisionTreeRegressor
from gplearn.genetic import SymbolicRegressor
from sklearn.kernel_ridge import KernelRidge
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct
from sklearn.gaussian_process.kernels import WhiteKernel
from xgboost import XGBRegressor
from maica.core.env import *
from maica.core.sys import *


class Model(ABC):
    """
    An abstract class of machine learning algorithm in MAICA.
    All machine learning algorithms in MAICA should inherit this class.
    """

    @abstractmethod
    def __init__(self,
                 alg_name: str):
        """
        :param alg_name: (*str*) A name of the algorithm defined in ``maica.core.env``.
        """

        self.alg_name = alg_name


class SKLearnModel(Model):
    """
    A wrapper class of the machine learning algorithms in the Scikit-learn library.
    This class provides a generic interface of ``fit()`` and ``predict()`` functions in the Scikit-learn algorithms.
    """

    def __init__(self,
                 alg_name: str,
                 **kwargs):
        """
        :param alg_name: (*str*) A name of the algorithm defined in ``maica.core.env``.
        :param kwargs: (*optional*) Hyper-parameters of the machine learning algorithms.
        """

        super(SKLearnModel, self).__init__(alg_name)
        self.alg = None

        # Initialize machine learning model with the given argument 'alg_name'.
        self.__init_model(**kwargs)

    def __init_model(self,
                     **kwargs):
        """
        Initialize machine learning model according to the class parameter ``self.alg_name``.

        :param kwargs: (*optional*) Hyper-parameters of the machine learning algorithms.
        """

        if self.alg_name == ALG_LR:
            # Linear Regression
            self.alg = LinearRegression()
        elif self.alg_name == ALG_LASSO:
            # LASSO
            alpha = kwargs['alpha'] if 'alpha' in kwargs.keys() else 0.1
            self.alg = Lasso(alpha=alpha)
        elif self.alg_name == ALG_DCTR:
            # Decision Tree Regression
            self.alg = DecisionTreeRegressor()
        elif self.alg_name == ALG_SYMR:
            # Symbolic Regression
            population_size = kwargs['population_size'] if 'population_size' in kwargs.keys() else 1000
            generations = kwargs['generations'] if 'generations' in kwargs.keys() else 100
            p_subtree_mutation = kwargs['p_subtree_mutation'] if 'p_subtree_mutation' in kwargs.keys() else 0.01
            p_hoist_mutation = kwargs['p_hoist_mutation'] if 'p_hoist_mutation' in kwargs.keys() else 0.01
            p_point_mutation = kwargs['p_point_mutation'] if 'p_point_mutation' in kwargs.keys() else 0.01
            self.alg = SymbolicRegressor(population_size=population_size, generations=generations,
                                         p_subtree_mutation=p_subtree_mutation, p_hoist_mutation=p_hoist_mutation,
                                         p_point_mutation=p_point_mutation, verbose=1)
        elif self.alg_name == ALG_KRR:
            # Kernel Ridge Regression
            alpha = kwargs['alpha'] if 'alpha' in kwargs.keys() else 1.0
            self.alg = KernelRidge(alpha=alpha)
        elif self.alg_name == ALG_KNNR:
            # K-Nearest Neighbor Regression
            n_neighbors = kwargs['n_neighbors'] if 'n_neighbors' in kwargs.keys() else 5
            self.alg = KNeighborsRegressor(n_neighbors=n_neighbors)
        elif self.alg_name == ALG_GPR:
            # Gaussian Process Regression
            kernel = DotProduct() + WhiteKernel()
            self.alg = GaussianProcessRegressor(kernel=kernel)
        elif self.alg_name == ALG_SVR:
            # Support Vector Regression
            c = kwargs['c'] if 'c' in kwargs.keys() else 1.0
            epsilon = kwargs['epsilon'] if 'epsilon' in kwargs.keys() else 0.1
            self.alg = SVR(C=c, epsilon=epsilon)
        elif self.alg_name == ALG_GBTR:
            # Gradient Boosting Tree Regression
            max_depth = kwargs['max_depth'] if 'max_depth' in kwargs.keys() else 7
            n_estimators = kwargs['n_estimators'] if 'n_estimators' in kwargs.keys() else 300
            self.alg = XGBRegressor(max_depth=max_depth, n_estimators=n_estimators)

    def fit(self,
            inputs: numpy.ndarray,
            targets: numpy.ndarray):
        """
        Fit model parameters for the given input and target data.

        :param inputs: (*numpy.ndarray*) The input data of the training dataset.
        :param targets: (*numpy.ndarray*) The target data of the training dataset.
        """

        self.alg.fit(inputs, targets)

    def predict(self,
                inputs: numpy.ndarray):
        """
        Predict target values of the given input data.

        :param inputs: (*numpy.ndarray*) The input data of the dataset.
        :return: (*numpy.ndarray*) Predicted values for the given ``inputs``.
        """

        return self.alg.predict(inputs)

    def save(self,
             path_model_file: str):
        """
        Save model parameters into a file in ``path_model_file``.

        :param path_model_file: (*str*) The path of the model file.
        """

        joblib.dump(self.alg, path_model_file)

    def load(self,
             path_model_file: str):
        """
        Load model parameters in a file of ``path_model_file``.

        :param path_model_file: (*str*) The path of the model file.
        """

        self.alg = joblib.load(path_model_file)


class PyTorchModel(torch.nn.Module, Model):
    """
    An abstract class of the machine learning algorithms in the PyTorch library.
    In MAICA, all PyTorch algorithms should inherit this class.
    """

    def __init__(self,
                 alg_name: str):
        """
        :param alg_name: (*str*) A name of the algorithm defined in ``maica.core.env``.
        """

        super(PyTorchModel, self).__init__()
        self.alg_type = ALG_PYTORCH
        self.alg_name = alg_name

    def gpu(self):
        """
        Move model parameters from CPU to GPU.

        :return: Model object (self) in GPU.
        """

        return self.cuda()

    def save(self,
             path_model_file: str):
        """
        Save model parameters into a file in ``path_model_file``.

        :param path_model_file: (*str*) The path of the model file.
        """

        torch.save(self.state_dict(), path_model_file)

    def load(self,
             path_model_file: str):
        """
        Load model parameters in a file of ``path_model_file``.

        :param path_model_file: (*str*) The path of the model file.
        """

        self.load_state_dict(torch.load(path_model_file))



"""
Crystal Structure
-----------------
This module provides data processing functions to convert the crystal structures into the mathematical graphs.
Like the molecular structure, the converted crystal structure are stored as a ``torch_geometric.Data`` object.
"""


import numpy
import torch
import pymatgen
from torch_geometric.data import Data
from pymatgen.core.structure import Structure
from sklearn.metrics import pairwise_distances
from maica_old.chem.base import atom_nums


def rbf(x: numpy.ndarray,
        mu: numpy.ndarray,
        beta: float):
    """
    Apply radial basis function to given data ``x`` with a given mean vector ``mu`` and a constant ``beta``.
    The Gaussian kernel was used to implement this radial basis function.

    :param x: (*numpy.ndarray*) An input data of radial basis function.
    :param mu: (*numpy.ndarray*) Feature-wise mean vector of the Gaussian kernel.
    :param beta: (*float*) A shape parameter of radial basis function.
    :return: (*numpy.ndarray*) Converted data of the input ``x``.
    """

    return numpy.exp(-(x - mu)**2 / beta**2)


def even_samples(min_val: float,
                 max_val: float,
                 n_samples: int):
    """
    Sample ``n_samples`` values evenly within ``[min_val, max_val]``.

    :param min_val: (*float*) The minimum value of the sample range.
    :param max_val: (*float*) The maximum value of the sample range.
    :param n_samples: (*int*) The number of samples.
    :return: (*numpy.ndarray*) Samples selected in the given range evenly.
    """

    samples = numpy.empty(n_samples)
    len_range = (max_val - min_val) / n_samples

    # Sample data evenly.
    for i in range(0, n_samples):
        samples[i] = min_val + len_range * (i + 1)

    return samples


def get_crystal_graph(elem_feats: numpy.ndarray,
                      path_cif: str,
                      numeric_feats: numpy.ndarray,
                      rbf_means: numpy.ndarray,
                      target: float = None,
                      gid: int = -1,
                      radius: float = 3.0):
    """
    Convert the crystallographic information framework (CIF) file into the mathematical graph
    :math:`G = (\mathcal{V}, \mathcal{E}, X, R)`, where :math:`\mathcal{V}` is a set of atoms,
    :math:`\mathcal{E}` is a set of bonds, :math:`X` is a atom-feature matrix,
    and :math:`R` is a bond-feature matrix.

    :param elem_feats: (*numpy.ndarray*) The NumPy array of elemental features.
    :param path_cif: (*str*) The path of the CIF file of the crystal structure.
    :param numeric_feats: (*numpy.ndarray*) Numerical features of the crystal (e.g., density).
    :param rbf_means: (*numpy.ndarray*) A feature-wise mean vector of the Gaussian kernel for radial basis function to generate bond features.
    :param target: (*float, optional*) The target property that you want to predict from the crystal structure (*default* = ``None``).
    :param gid: (*int, optional*) An integer identifier of the crystal structure (*default* = -1).
    :param radius: (*float, optional*) The maximum value of the radius to define neighbor atoms (*default* = 3.0).
    :return: (*torch_geometric.Data*) A crystal graph :math:`G = (\mathcal{V}, \mathcal{E}, X, R)`.
    """

    try:
        # Get Crystal object from the given CIF.
        crystal = Structure.from_file(path_cif)

        # Get the atomic information of the crystal structure.
        atom_coord, atom_feats = get_atom_info(crystal, elem_feats, numeric_feats, radius)

        # Get the bond information of the crystal structure.
        bonds, bond_feats = get_bond_info(atom_coord, rbf_means, radius)

        # Check isolated graph.
        if bonds is None:
            return None

        # Convert numpy.ndarray to torch.tensor.
        atom_feats = torch.tensor(atom_feats, dtype=torch.float)
        bonds = torch.tensor(bonds, dtype=torch.long).t().contiguous()
        bond_feats = torch.tensor(bond_feats, dtype=torch.float)
        gid = torch.tensor(gid, dtype=torch.long).view(1, 1)

        # Check target property.
        if target is not None:
            target = torch.tensor(target, dtype=torch.float).view(1, 1)

        # Return a generated crystal graph as torch_geometric.Data object.
        return Data(x=atom_feats, edge_index=bonds, edge_attr=bond_feats, y=target, gid=gid)
    except RuntimeError:
        return None


def get_atom_info(crystal: pymatgen.core.Structure,
                  elem_feats: numpy.ndarray,
                  numeric_feats: numpy.ndarray,
                  radius: float):
    """
    Calculate an atom-feature matrix and an atom-coordination matrix from the given ``pymatgen.core.Structure`` object.

    :param crystal: (*pymatgen.core.Structure*) The Pymatgen object of the crystal structure.
    :param elem_feats: (*numpy.ndarray*) The NumPy array of elemental features.
    :param numeric_feats: (*numpy.ndarray*) Numerical features of the crystal (e.g., density).
    :param radius: (*float*) The maximum value of the radius to define neighbor atoms.
    :return: (*numpy.ndarray, numpy.ndarray*) The atom-feature and atom-coordination matrices.
    """

    atoms = list(crystal.atomic_numbers)
    atom_coord = list()
    atom_feats = list()
    list_nbrs = crystal.get_all_neighbors(radius)

    # Get overall charge of the crystal structure.
    charge = crystal.charge

    # Get density in units of g/cc.
    density = float(crystal.density)

    # Get volume of the crystal structure.
    volume = crystal.volume

    coords = dict()
    for coord in list(crystal.cart_coords):
        coord_key = ','.join(list(coord.astype(str)))
        coords[coord_key] = True

    for i in range(0, len(list_nbrs)):
        nbrs = list_nbrs[i]

        for j in range(0, len(nbrs)):
            coord_key = ','.join(list(nbrs[j][0].coords.astype(str)))
            if coord_key not in coords.keys():
                coords[coord_key] = True
                atoms.append(atom_nums[nbrs[j][0].species_string])

    for coord in coords.keys():
        atom_coord.append(numpy.array([float(x) for x in coord.split(',')]))
    atom_coord = numpy.vstack(atom_coord)

    for i in range(0, len(atoms)):
        # Get elemental attributes of the atom.
        elem_attr = elem_feats[atoms[i] - 1, :]

        # Concatenate the given numerical features if they exist.
        if numeric_feats is None:
            atom_feats.append(numpy.hstack([elem_attr, charge, density, volume]))
        else:
            atom_feats.append(numpy.hstack([elem_attr, charge, density, volume, numeric_feats]))
    atom_feats = numpy.vstack(atom_feats).astype(float)

    return atom_coord, atom_feats


def get_bond_info(atom_coord: numpy.ndarray,
                  rbf_means: numpy.ndarray,
                  radius: float):
    """
    Calculate bond information of the crystal structure.
    If the crystal structure was converted into an isolated graph, it returns (None, None).

    :param atom_coord: (*numpy.ndarray*) The XYZ coordinates of the atoms in the crystal structure.
    :param rbf_means: (*numpy.ndarray*) A feature-wise mean vector of the Gaussian kernel for radial basis function to generate bond features.
    :param radius: (*float*) The maximum value of the radius to define neighbor atoms.
    :return: (*numpy.ndarray, numpy.ndarray*) Indices of the bonds in the crystal structure and the bond-feature matrix.
    """

    bonds = list()
    bond_feats = list()
    pdist = pairwise_distances(atom_coord)

    # Calculate bond information.
    for i in range(0, atom_coord.shape[0]):
        for j in range(0, atom_coord.shape[0]):
            if i != j and pdist[i, j] <= radius:
                bonds.append([i, j])
                bond_feats.append(rbf(numpy.full(rbf_means.shape[0], pdist[i, j]), rbf_means, beta=0.5))

    if len(bonds) == 0:
        # If there is no bond in the given crystal structure.
        return None, None
    else:
        bonds = numpy.vstack(bonds)
        bond_feats = numpy.vstack(bond_feats)

        return bonds, bond_feats



"""
Visualization
-------------
The ``maica_old.util.visualization`` module provides various visualization tools for dataset analysis and evaluations.
This module was implemented based on ``matplotlib`` to serve a unified interface.
"""


import numpy
import matplotlib.pyplot as plt
from sklearn.manifold import TSNE
from sklearn.manifold import SpectralEmbedding
from maica_old.core.env import *
from maica_old.data.base import Dataset
from maica_old.util.analysis import get_error_dist


def plot_target_dist(fig_name: str,
                     dataset: Dataset,
                     n_bins: int = 10,
                     font_size: int = 16):
    """
    Plot a histogram of the target values of ``dataset``.
    The histogram is saved as an image file in ``fig_name``.

    :param fig_name: (*str*) A name of the image file storing the histogram of the target values.
    :param dataset: (*maica_old.data.base.Dataset*) The dataset object containing the target values.
    :param n_bins: (*int, optional*) The number of bins in the histogram (*default* = 10).
    :param font_size: (*int, optional*) The size of text in the generated figure (*default* = 16).
    """

    targets = dataset.y
    y_min = numpy.min(targets)
    y_max = numpy.max(targets)
    label_bins = numpy.linspace(y_min, y_max, n_bins + 1)

    plt.tight_layout()
    plt.rcParams.update({'font.size': font_size})
    plt.title('Histogram of Target Distribution')
    plt.xlabel('Target range')
    plt.ylabel('# of data')
    plt.hist(targets, label_bins, edgecolor='black', zorder=2)
    plt.grid(linestyle='--')
    plt.savefig(fig_name, bbox_inches='tight', dpi=500)
    plt.close()


def plot_error_dist(fig_name: str,
                    dataset: Dataset,
                    preds: numpy.ndarray,
                    font_size: int = 16):
    """
    Plot a distribution of prediction errors.
    The prediction errors for each target range are calculated based on mean absolute error as:

    .. math::
        error = \\frac{1}{N_i}\sum_{j=1}^{N_i}|y_j - y_j^{'}|,

    where :math:`N_i` is the number of data points in the :math:`i^{th}` target range,
    :math:`y_j` is the target value of the :math:`j^{th}` data in the :math:`i^{th}` target range,
    and :math:`y_{j}^{'}` is the predicted value of the :math:`j^{th}` data.

    :param fig_name: (*str*) A name of the image file storing the error distribution.
    :param dataset: (*maica_old.data.base.Dataset*) The dataset object containing the target values.
    :param preds: (*numpy.ndarray*) Prediction results.
    :param font_size: (*int, optional*) The size of text in the generated figure (*default* = 16).
    """

    errors, labels = get_error_dist(dataset.y, preds)
    labels = [(labels[i] + labels[i + 1]) / 2 for i in range(0, len(labels) - 1)]
    width_bar = (labels[1] - labels[0])

    plt.tight_layout()
    plt.rcParams.update({'font.size': font_size})
    plt.title('Error Distribution for Each Target range')
    plt.xlabel('Target range')
    plt.ylabel('Mean of prediction errors')
    plt.bar(labels, errors, edgecolor='black', width=width_bar, zorder=2)
    plt.grid(linestyle='--')
    plt.savefig(fig_name, bbox_inches='tight', dpi=500)
    plt.close()


def plot_pred_result(fig_name: str,
                     dataset: object,
                     preds: numpy.ndarray,
                     font_size: int = 16,
                     min_val: float = None,
                     max_val: float = None):
    """
    Draw a scatter plot of the prediction results in ``preds``.
    The X and Y axes are drawn by the true target values in ``dataset`` and the predicted values in ``preds``, respectively.

    :param fig_name: (*str*) A name of the image file storing the scatter plot.
    :param dataset: (*str*) The dataset object containing the target values.
    :param preds: (*numpy.ndarray*) Prediction results.
    :param font_size: (*int, optional*) The size of text in the generated figure (*default* = 16).
    :param min_val: (*float, optional*) The minimum value of the X and Y axes (*default* = ``None``).
    :param max_val: (*float, optional*) The maximum value of the X and Y axes (*default* = ``None``).
    """

    targets = dataset.y

    if min_val is None:
        min_val = numpy.minimum(numpy.min(targets), numpy.min(preds))
        min_val -= 0.2 * min_val

    if max_val is None:
        max_val = numpy.maximum(numpy.max(targets), numpy.max(preds))
        max_val += 0.2 * max_val

    plt.tight_layout()
    plt.rcParams.update({'font.size': font_size})
    plt.title('Scatter Plot of Prediction Results')
    plt.xlabel('Target value')
    plt.ylabel('Predicted value')
    plt.grid(linestyle='--')
    plt.xlim([min_val, max_val])
    plt.ylim([min_val, max_val])
    plt.plot([min_val, max_val], [min_val, max_val], 'k', zorder=2)
    plt.scatter(targets, preds, edgecolor='k', zorder=3)
    plt.savefig(fig_name, bbox_inches='tight', dpi=500)
    plt.close()


def plot_embeddings(fig_name: str,
                    data: numpy.ndarray,
                    labels: numpy.ndarray,
                    font_size: int = 16,
                    emb_method: str = EMB_TSNE):
    """
    Draw a scatter plot of the data embeddings from ``data`` and ``emb_method``.
    This function generates 2-dimensional scatter plot based on first and second components of the embedding method.
    Available embedding method are given in ``maica_old.core.env``.

    :param fig_name: (*str*) A name of the image file storing the scatter plot.
    :param data: (*numpy.ndarray*) The original data to be embedded.
    :param labels: (*labels*) Numerical labels for the data.
    :param font_size: (*int, optional*) The size of text in the generated figure (*default* = 16).
    :param emb_method: (*str, optional*) Embedding method (*default* = ``env.EMB_TSNE``).
    """

    if emb_method == EMB_TSNE:
        embs = TSNE(n_components=2).fit_transform(data)
    elif emb_method == EMB_SPECT:
        embs = SpectralEmbedding(n_components=2).fit_transform(data)

    plt.tight_layout()
    plt.rcParams.update({'font.size': font_size})
    plt.title('Scatter Plot of Data Embeddings')
    plt.xlabel('Latent Feature #1')
    plt.ylabel('Latent Feature #2')
    plt.scatter(embs[:, 0], embs[:, 1], edgecolor='k', c=labels, zorder=3)
    plt.colorbar()
    plt.savefig(fig_name, bbox_inches='tight', dpi=500)
    plt.close()
